from angrmanagement.ui.icons import icon

from .menu import Menu, MenuEntry


class PluginMenu(Menu):
    def __init__(self, main_window):
        super().__init__("&Plugins", parent=main_window)

        self.entries.extend(
            [MenuEntry("&Manage Plugins...", main_window.open_load_plugins_dialog, icon=icon("plugins"))]
        )



from typing import TYPE_CHECKING

from PySide6.QtWidgets import QFontDialog, QPushButton

from angrmanagement.config import Conf

if TYPE_CHECKING:
    from PySide6.QtGui import QFont


class QFontOption(QPushButton):
    """
    A widget used to allow users to change a font stored in Conf
    """

    def __init__(self, config_key: str, parent=None):
        """
        :param key: Key of the font in Conf
        :parent:    Optional parent of this QWidget
        """
        super().__init__(parent)
        self._config_key = config_key
        self.font: QFont = getattr(Conf, config_key)
        self.released.connect(self._prompt)
        self._format()

    def update(self):
        """
        Update Conf with the selected font
        """
        setattr(Conf, self._config_key, self.font)

    def _format(self):
        """
        Format the widget's UI elements
        """
        self.setText(f"{self.font.family()}, {self.font.pointSize()} pt")
        self.setFont(self.font)  # We do not ._sync.keep_synced this; it should update always

    def _prompt(self):
        """
        Prompt the user to select a font
        """
        ok, font = QFontDialog.getFont(self.font, parent=self)
        if ok:
            self.font = font
            self._format()



"""
monkeypatch_stdio - Monkeypatches stdout and stderr to be NullWriters.
"""

import sys


class NullWriter:  # pylint: disable=no-self-use,unused-argument
    """
    A file-like object that does nothing.
    """

    softspace = 0
    encoding = "UTF-8"

    def write(self, *args):
        pass

    def flush(self, *args):
        pass

    def isatty(self):
        return False


def monkeypatch_stdio():
    """
    Monkeypatch stdout and stderr to be NullWriters.
    """
    if sys.stdout is None:
        sys.stdout = NullWriter()
    if sys.stderr is None:
        sys.stderr = NullWriter()



from typing import Optional

from angrmanagement.ui.views.disassembly_view import DisassemblyView


class DiffDisassemblyView(DisassemblyView):
    """
    A Disassembly View for a binary being Diffed. Should never try to synchronize normally since
    it will almost certainly have different addresses
    """

    def on_synchronized_cursor_address_changed(self):
        assert not self._processing_synchronized_cursor_update
        self._processing_synchronized_cursor_update = True
        try:
            if self.sync_state.cursor_address is not None:
                self.instance.recompilation_plugin.syncronize_with_original_disassembly_view()
        finally:
            self._processing_synchronized_cursor_update = False

    def set_synchronized_cursor_address(self, address: Optional[int]):
        pass



import os
import typing

from angr.storage.file import SimFileDescriptor
from PySide6.QtWidgets import QComboBox, QFileDialog, QFrame, QPushButton, QTextEdit, QVBoxLayout

if typing.TYPE_CHECKING:
    from angr.sim_state import SimState


class QFileDescriptorViewer(QFrame):
    """
    embeded in `StateInspector`
    Display content of a file descriptor ( include stdin/stdout/stderr ) for the selected state.
    """

    STD = {
        0: "stdin",
        1: "stdout",
        2: "stderr",
    }

    def __init__(self, state, parent, workspace):
        super().__init__(parent)

        self._state: SimState = state
        self.workspace = workspace

        self.select_fd = None
        self._current_fd = None
        self.textedit = None

        self._state.am_subscribe(self._watch_state)

    def dump_fd(self, fd):
        # Clean up when nothing is selected
        if fd == -1:
            self._current_fd = None
            self.textedit.setPlainText("")
            return

        if self._state.am_none:
            return
        self._current_fd = fd
        self.textedit.setPlainText(self._state.posix.dumps(fd).decode("ascii", "ignore").replace("\x00", "\\x00"))

    def save_as(self):
        if self._state.am_none or self._current_fd is None:
            return
        filename, folder = QFileDialog.getSaveFileName(self, "Save content to ...", "", "Any file (*);")
        if filename and folder:
            save_to = os.path.join(folder, filename)
            open(save_to, "wb").write(self._state.posix.dumps(self._current_fd))

    def _init_widgets(self):
        layout = QVBoxLayout()
        self.select_fd = QComboBox(self)
        self.select_fd.currentIndexChanged.connect(self.dump_fd)
        layout.addWidget(self.select_fd)

        self.textedit = QTextEdit(self)
        self.textedit.setAcceptRichText(False)
        self.textedit.setReadOnly(True)
        layout.addWidget(self.textedit)

        save_as = QPushButton("Save as...", self)
        save_as.clicked.connect(self.save_as)
        layout.addWidget(save_as)

        self.setLayout(layout)

    def _watch_state(self, **kwargs):  # pylint: disable=unused-argument
        if self._state.am_none:
            return
        if self.select_fd is None:
            self._init_widgets()
        self.select_fd.clear()
        for fd, simfile in self._state.posix.fd.items():
            if fd in self.STD:
                self.select_fd.addItem(self.STD[fd])
            elif isinstance(simfile, SimFileDescriptor):
                self.select_fd.addItem(str(simfile.file.name))
            else:
                self.select_fd.addItem(str(simfile))



class EdgeSort:
    DIRECT_JUMP = 0
    TRUE_BRANCH = 1
    FALSE_BRANCH = 2
    BACK_EDGE = 3
    EXCEPTION_EDGE = 4


class Edge:
    def __init__(self, src, dst, sort=EdgeSort.DIRECT_JUMP):
        self.src = src
        self.dst = dst

        self.start_index = None
        self.max_start_index = None
        self.end_index = None
        self.max_end_index = None

        self.points = []
        self.moves = []
        self.coordinates = []
        self.sort = sort

    def add_point(self, col, row, index):
        self.points.append((col, row, index))

    def add_move(self, move):
        self.moves.append(move)

    def add_coordinate(self, x, y):
        if len(self.coordinates) >= 2:
            coord_a, coord_b = self.coordinates[-2], self.coordinates[-1]
            if coord_b[0] == coord_a[0] == x:
                # it moves vertically
                # replace coord_b
                self.coordinates[-1] = (x, y)
                return
            elif coord_b[1] == coord_a[1] == y:
                # it moves horizontally
                # replace coord b
                self.coordinates[-1] = (x, y)
                return

        self.coordinates.append((x, y))

    @property
    def first_move(self):
        if self.moves:
            return self.moves[0]

        return 1  # NO_MOVE

    @property
    def last_move(self):
        if self.moves:
            return self.moves[-1]

        return 1  # NO_MOVE

    def __repr__(self):
        return "<Edge between %s and %s, %d coordinates>" % (self.src, self.dst, len(self.coordinates))



import platform

from PySide6.QtGui import QAction, QKeySequence

from angrmanagement.logic import GlobalInfo
from angrmanagement.ui.icons import icon

from .menu import Menu, MenuEntry, MenuSeparator

try:
    import archr
except ImportError:
    archr = None


class RecentMenuEntry(MenuEntry):
    """
    Represents an entry in the "Load recent" list. Holds a path and an indication of what's at that path.
    """

    def __init__(self, path):
        self.path = path
        super().__init__(path, self.action_target)

    def action_target(self):
        GlobalInfo.main_window.load_file(self.path)


class FileMenu(Menu):
    """
    Lays out the entries under the 'File' menu
    """

    def __init__(self, main_window):
        super().__init__("&File", parent=main_window)
        self._project = main_window.workspace.main_instance.project

        self._save_entries = [
            MenuEntry("&Save angr database...", main_window.save_database, shortcut=QKeySequence("Ctrl+S")),
            MenuEntry("S&ave angr database as...", main_window.save_database_as, shortcut=QKeySequence("Ctrl+Shift+S")),
            MenuEntry("Save patched binary as...", main_window.save_patched_binary_as),
        ]
        self._edit_save()
        self._project.am_subscribe(self._edit_save)

        self.recent_menu = Menu("Load recent")
        self.entries.extend(
            [
                MenuEntry("L&oad a new binary...", main_window.open_file_button, shortcut=QKeySequence("Ctrl+O")),
                *(
                    []
                    if archr is None
                    else [
                        MenuEntry(
                            "Loa&d a new docker target...",
                            main_window.open_docker_button,
                            shortcut=QKeySequence("Ctrl+Shift+O"),
                        ),
                    ]
                ),
                MenuEntry(
                    "Load a &trace file...",
                    main_window.open_trace_file_button,
                    shortcut=QKeySequence("Ctrl+Shift+T"),
                ),
                self.recent_menu,
                MenuSeparator(),
                MenuEntry("&Load angr database...", main_window.load_database, shortcut=QKeySequence("Ctrl+L")),
                *self._save_entries,
                MenuSeparator(),
                MenuEntry("Load a new &trace...", main_window.load_trace),
                MenuSeparator(),
                MenuEntry(
                    "&Preferences...",
                    main_window.preferences,
                    shortcut=QKeySequence("Ctrl+Comma"),
                    role=QAction.PreferencesRole,
                    icon=icon("preferences") if platform.system() != "Darwin" else None,
                ),
                MenuSeparator(),
                MenuEntry("E&xit", main_window.quit),
            ]
        )

    def _edit_save(self, **_):
        enable: bool = not self._project.am_none
        for i in self._save_entries:
            (i.enable if enable else i.disable)()

    def add_recent(self, path: str):
        for entry in list(self.recent_menu.entries):
            if entry.path == path:
                self.recent_menu.remove(entry)
        self.recent_menu.add(RecentMenuEntry(path), 0)



from typing import TYPE_CHECKING, List

from angr import options
from angr.state_plugins.heap import SimHeapPTMalloc
from sortedcontainers.sorteddict import SortedDict

from angrmanagement.plugins.base_plugin import BasePlugin

if TYPE_CHECKING:
    from angr.sim_state import SimState
    from angr.state_plugins.sim_action import SimAction


class MemoryChecker(BasePlugin):
    AllowList = ["free", "malloc", "__libc_start_main"]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.states = self.workspace.main_instance.states
        self.states.am_subscribe(self.install_state_plugin)

    def install_state_plugin(self, **kwargs):
        if kwargs.get("src", None) != "new":
            return
        state: SimState = kwargs.get("state")
        state.register_plugin("heap", SimHeapPTMalloc())
        state.options.update({options.TRACK_MEMORY_ACTIONS})

    @staticmethod
    def eval_ptr(state, ptr):
        return state.solver.eval(ptr)

    @staticmethod
    def check_address_is_free(state: "SimState", ptr_list: "List[SimAction]"):
        ptr_dict = SortedDict([(MemoryChecker.eval_ptr(state, x.addr.ast), x) for x in ptr_list])
        len_list = len(ptr_dict)
        for chunk in state.heap.free_chunks():
            base = chunk.base
            size = state.solver.eval(chunk.get_size())
            p = ptr_dict.bisect_left(base)
            if p < len_list and base <= ptr_dict.peekitem(p)[0] < base + size:
                if state.posix.stderr.writable:
                    addr, act = ptr_dict.peekitem(p)
                    err_str = (
                        f"\n=== Use-After-Free Plugin ===\n"
                        f"Memory Address:{addr:#x}\n"
                        f"Instrument Address:{act.ins_addr:#x}\n"
                    )
                    state.posix.stderr.write(None, err_str.encode())
                return True
        return False

    @staticmethod
    def check_use_after_free(state: "SimState"):
        # heap: SimHeapPTMalloc = state.heap
        # heap.print_heap_state()

        actions: List[SimAction] = state.history.actions.hardcopy
        last_bbl_addr = actions[-1].bbl_addr
        address_list = []
        for act in reversed(actions):
            if act.bbl_addr != last_bbl_addr:
                break
            if act.type == "mem" and (
                act.sim_procedure is None or act.sim_procedure.display_name not in MemoryChecker.AllowList
            ):
                address_list.append(act)
        return MemoryChecker.check_address_is_free(state, address_list)

    def step_callback(self, simgr):
        simgr.move("active", "use_after_free", self.check_use_after_free)



from PySide6.QtWidgets import QAbstractItemView, QFileDialog, QHeaderView, QMenu, QTableWidget, QTableWidgetItem


class QFileSystemTable(QTableWidget):
    def __init__(self, items, parent):
        super().__init__(parent)

        header_labels = ["Mount Point", "Host Path"]

        self.setColumnCount(len(header_labels))
        self.setHorizontalHeaderLabels(header_labels)
        self.setSelectionBehavior(QAbstractItemView.SelectItems)
        header = self.horizontalHeader()
        header.setSectionResizeMode(0, QHeaderView.Stretch)
        header.setSectionResizeMode(1, QHeaderView.Stretch)

        self.setRowCount(len(items))
        for idx, item in enumerate(items):
            for i, it in enumerate(item):
                self.setItem(idx, i, QTableWidgetItem(it))

    def contextMenuEvent(self, event):
        sr = self.currentRow()

        menu = QMenu("", self)

        menu.addAction("Add a Row", self._action_new_row)
        menu.addSeparator()

        a = menu.addAction("Delete this Row", self._action_delete)
        if sr is None:
            a.setDisabled(True)
        b = menu.addAction("Select a File", self._action_select_file)
        if sr is None:
            b.setDisabled(True)
        c = menu.addAction("Select a directory", self._action_select_dir)
        if sr is None:
            c.setDisabled(True)

        menu.exec_(event.globalPos())

    def _action_new_row(self):
        row = self.rowCount()
        self.insertRow(row)
        self.setItem(row, 0, QTableWidgetItem("Edit Me"))
        self.setItem(row, 1, QTableWidgetItem(""))

    def _action_select_file(self):
        file_path, succ = QFileDialog.getOpenFileName(
            self,
            "Open a real file",
            "",
            "All executables (*)",
        )
        if succ:
            self.setItem(self.currentRow(), 1, QTableWidgetItem(file_path))

    def _action_select_dir(self):
        dir_path = QFileDialog.getExistingDirectory(self, "Select a directory")
        if dir_path:
            self.setItem(self.currentRow(), 1, QTableWidgetItem(dir_path))

    def _action_delete(self):
        self.removeRow(self.currentRow())

    def get_result(self):
        ret = []
        for i in range(self.rowCount()):
            ret.append([self.item(i, 0).text(), self.item(i, 1).text()])
        return ret



from typing import TYPE_CHECKING, Mapping, Type

from PySide6.QtCore import Qt
from PySide6.QtWidgets import QDockWidget

from angrmanagement.ui.toolbars import DebugToolbar, FeatureMapToolbar, FileToolbar

if TYPE_CHECKING:
    from angrmanagement.ui.main_window import MainWindow
    from angrmanagement.ui.toolbars.toolbar import Toolbar


class ToolbarManager:
    """
    Manages toolbars shown on the main window.
    """

    def __init__(self, main_window):
        self._main_window: MainWindow = main_window
        self.active: Mapping[Type[Toolbar], Toolbar] = {}
        self.all_toolbars = [FileToolbar, DebugToolbar, FeatureMapToolbar]

    @staticmethod
    def get_name_for_toolbar_class(toolbar_cls: Type["Toolbar"]) -> str:
        return {FileToolbar: "File", DebugToolbar: "Debug", FeatureMapToolbar: "Feature Map"}[toolbar_cls]

    def show_toolbar_by_class(self, cls: Type["Toolbar"]):
        if cls not in self.active:
            tb = cls(self._main_window)
            self.active[cls] = tb
            qtb = tb.qtoolbar()
            if isinstance(qtb, QDockWidget):
                self._main_window.addDockWidget(Qt.TopDockWidgetArea, qtb)
            else:
                self._main_window.addToolBar(Qt.TopToolBarArea, qtb)
        else:
            self.active[cls].qtoolbar().show()

    def hide_toolbar_by_class(self, cls: Type["Toolbar"]):
        if cls in self.active:
            tb = self.active.pop(cls)
            qtb = tb.qtoolbar()
            if isinstance(qtb, QDockWidget):
                self._main_window.removeDockWidget(qtb)
            else:
                self._main_window.removeToolBar(qtb)
            tb.shutdown()

    def set_toolbar_visible_by_class(self, cls: Type["Toolbar"], visible: bool):
        if visible:
            self.show_toolbar_by_class(cls)
        else:
            self.hide_toolbar_by_class(cls)

    def show_all(self):
        for cls in self.all_toolbars:
            self.show_toolbar_by_class(cls)

    def hide_all(self):
        for cls in self.all_toolbars:
            self.hide_toolbar_by_class(cls)



import logging
import time

from angrmanagement.data.analysis_options import CFGForceScanMode
from angrmanagement.logic.threads import gui_thread_schedule_async

from .job import Job

_l = logging.getLogger(name=__name__)


class CFGGenerationJob(Job):
    """
    Job for generating the Control Flow Graph.
    """

    DEFAULT_CFG_ARGS = {
        "normalize": True,  # this is what people naturally expect
    }

    def __init__(self, on_finish=None, **kwargs):
        super().__init__(name="CFG generation", on_finish=on_finish)

        # TODO: sanitize arguments

        # make a copy
        cfg_args = dict(kwargs)
        for key, val in self.DEFAULT_CFG_ARGS.items():
            if key not in cfg_args:
                cfg_args[key] = val

        self.cfg_args = cfg_args

        scanning_mode = self.cfg_args.pop("scanning_mode", None)
        if scanning_mode is not None:
            self.cfg_args["force_smart_scan"] = scanning_mode == CFGForceScanMode.SmartScan
            self.cfg_args["force_complete_scan"] = scanning_mode == CFGForceScanMode.CompleteScan

        self._cfb = None
        self._last_progress_callback_triggered = None
        self.instance = None

    def _run(self, inst):
        self.instance = inst
        exclude_region_types = {"kernel", "tls"}
        # create a temporary CFB for displaying partially analyzed binary during CFG recovery
        temp_cfb = inst.project.analyses.CFB(
            exclude_region_types=exclude_region_types, on_object_added=self._on_cfb_object_added
        )
        self._cfb = temp_cfb

        cfg = inst.project.analyses.CFG(
            progress_callback=self._progress_callback,
            low_priority=True,
            cfb=temp_cfb,
            **self.cfg_args,
        )
        self._cfb = None
        # Build the real one
        cfb = inst.project.analyses.CFB(kb=cfg.kb, exclude_region_types=exclude_region_types)

        return cfg, cfb

    def finish(self, inst, result):
        try:
            cfg, cfb = result
            inst.cfb = cfb
            inst.cfg = cfg.model
            inst.cfb.am_event()
            inst.cfg.am_event()
            super().finish(inst, result)
        except Exception:  # pylint:disable=broad-exception-caught
            _l.error("Exception occurred in CFGGenerationJob.finish().", exc_info=True)

    def __repr__(self):
        return "Generating CFG"

    #
    # Private methods
    #

    def _progress_callback(self, percentage, text=None, cfg=None):
        t = time.time()
        if self._last_progress_callback_triggered is not None and t - self._last_progress_callback_triggered < 0.2:
            return
        self._last_progress_callback_triggered = t
        super()._progress_callback(percentage, text=text)

        if cfg is not None:
            # Peek into the CFG
            gui_thread_schedule_async(
                self._refresh,
                args=(
                    cfg,
                    self._cfb,
                ),
            )

    def _refresh(self, cfg, cfb):
        # do not signal events. that will happen on a timer to not overwhelm the renderer
        # instance will exist because _run must be used first
        self.instance.cfg = cfg
        self.instance.cfb = cfb

    def _on_cfb_object_added(self, addr, obj):
        self.instance.cfb.am_event(object_added=(addr, obj))



import os
import string
from functools import partial
from typing import TYPE_CHECKING, List, Optional, Set, Tuple

import qtawesome as qta
from angr.analyses.code_tagging import CodeTags
from cle.backends.uefi_firmware import UefiPE
from PySide6.QtCore import SIGNAL, QAbstractTableModel, QEvent, Qt
from PySide6.QtGui import QAction, QBrush, QColor, QCursor
from PySide6.QtWidgets import (
    QAbstractItemView,
    QHBoxLayout,
    QHeaderView,
    QLabel,
    QLineEdit,
    QMenu,
    QTableView,
    QVBoxLayout,
    QWidget,
)

from angrmanagement.config import Conf
from angrmanagement.data.instance import ObjectContainer
from angrmanagement.ui.menus.function_context_menu import FunctionContextMenu
from angrmanagement.ui.toolbars import FunctionTableToolbar

if TYPE_CHECKING:
    import PySide6
    from angr.knowledge_plugins.functions import Function, FunctionManager

    from angrmanagement.ui.views.functions_view import FunctionsView


class QFunctionTableModel(QAbstractTableModel):
    """
    The table model for QFunctionTable.
    """

    Headers = ["Name", "Tags", "Address", "Binary", "Size", "Blocks", "Complexity"]
    NAME_COL = 0
    TAGS_COL = 1
    ADDRESS_COL = 2
    BINARY_COL = 3
    SIZE_COL = 4
    BLOCKS_COL = 5
    COMPLEXITY_COL = 6

    def __init__(self, workspace, instance, func_list):
        super().__init__()

        self._func_list = None
        self._raw_func_list = func_list
        self.workspace = workspace
        self.instance = instance
        self._config = Conf
        self._data_cache = {}

    def __len__(self):
        if self._func_list is not None:
            return len(self._func_list)
        if self._raw_func_list is not None:
            return len(self._raw_func_list)
        return 0

    @property
    def func_list(self) -> List["Function"]:
        if self._func_list is not None:
            return self._func_list
        return self._raw_func_list

    @func_list.setter
    def func_list(self, v):
        self._func_list = None
        self._raw_func_list = v
        self._data_cache.clear()
        self.emit(SIGNAL("layoutChanged()"))

    def filter(self, keyword):
        if not keyword or self._raw_func_list is None:
            # remove the filtering
            self._func_list = None
        else:
            extra_columns = self.workspace.plugins.count_func_columns()
            self._func_list = [
                func
                for func in self._raw_func_list
                if self._func_match_keyword(func, keyword, extra_columns=extra_columns)
            ]

        self._data_cache.clear()
        self.emit(SIGNAL("layoutChanged()"))

    def rowCount(self, *args, **kwargs):  # pylint:disable=unused-argument
        if self.func_list is None:
            return 0
        return len(self.func_list)

    def columnCount(self, *args, **kwargs):  # pylint:disable=unused-argument
        return len(self.Headers) + self.workspace.plugins.count_func_columns()

    def headerData(self, section, orientation, role):  # pylint:disable=unused-argument
        if role != Qt.DisplayRole:
            return None

        if section < len(self.Headers):
            return self.Headers[section]
        else:
            try:
                return self.workspace.plugins.get_func_column(section - len(self.Headers))
            except IndexError:
                # Not enough columns
                return None

    def data(self, index, role):
        if not index.isValid():
            return None

        row = index.row()
        if row >= len(self):
            return None

        col = index.column()

        key = (row, col, role)
        if key in self._data_cache:
            value = self._data_cache[key]
        else:
            value = self._data_uncached(row, col, role)
            self._data_cache[key] = value
        return value

    def _data_uncached(self, row, col, role):
        func = self.func_list[row]

        if role == Qt.DisplayRole:
            return self._get_column_text(func, col)

        elif role == Qt.ForegroundRole:
            if func.is_syscall:
                color = self._config.function_table_syscall_color
            elif func.is_plt:
                color = self._config.function_table_plt_color
            elif func.is_simprocedure:
                color = self._config.function_table_simprocedure_color
            elif func.alignment:
                color = self._config.function_table_alignment_color
            else:
                color = self._config.function_table_color

            # for w in widgets:
            #    w.setFlags(w.flags() & ~Qt.ItemIsEditable)
            #    w.setForeground(color)

            return QBrush(color)

        elif role == Qt.BackgroundColorRole:
            color = self.workspace.plugins.color_func(func)
            if color is None:
                # default colors
                if func.from_signature:
                    color = self._config.function_table_signature_bg_color
            return color

        elif role == Qt.FontRole:
            return Conf.tabular_view_font

        return None

    def sort(self, column, order):
        self.layoutAboutToBeChanged.emit()
        self.func_list = sorted(
            self.func_list, key=lambda f: self._get_column_data(f, column), reverse=order == Qt.DescendingOrder
        )
        self.layoutChanged.emit()

    #
    # Private methods
    #

    def _get_column_data(self, func, idx):
        if idx == self.NAME_COL:
            return func.demangled_name
        elif idx == self.TAGS_COL:
            return func.tags
        elif idx == self.ADDRESS_COL:
            return func.addr
        elif idx == self.BINARY_COL:
            return self._get_binary_name(func)
        elif idx == self.SIZE_COL:
            return func.size
        elif idx == self.BLOCKS_COL:
            return len(func.block_addrs_set)
        elif idx == self.COMPLEXITY_COL:
            return func.cyclomatic_complexity
        else:
            return self.workspace.plugins.extract_func_column(func, idx - len(self.Headers))[0]

    def _get_column_text(self, func, idx):
        if idx < len(self.Headers):
            data = self._get_column_data(func, idx)
            if idx == self.ADDRESS_COL:
                return f"{data:x}"
            elif idx == self.TAGS_COL:
                return self._get_tags_display_string(data)
            else:
                return str(data)

        return self.workspace.plugins.extract_func_column(func, idx - len(self.Headers))[1]

    @staticmethod
    def _get_binary_name(func) -> str:
        if func.binary is not None:
            if func.binary.binary is not None:
                return os.path.basename(func.binary.binary)
            if isinstance(func.binary, UefiPE):
                if func.binary.user_interface_name:
                    return func.binary.user_interface_name
                if func.binary.guid:
                    return str(func.binary.guid)
                return str(func.binary)
        return ""

    def _get_function_backcolor(self, func) -> QColor:
        return self._backcolor_callback(func)

    TAG_STRS = {
        CodeTags.HAS_XOR: "Xor",
        CodeTags.HAS_BITSHIFTS: "Shift",
        CodeTags.HAS_SQL: "SQL",
    }

    @classmethod
    def _get_tags_display_string(cls, tags):
        return ", ".join(cls.TAG_STRS.get(t, t) for t in tags)

    def _func_match_keyword(self, func, keyword, extra_columns: int = 0):
        """
        Check whether the function matches against the given keyword or not.

        :param func:        The function to match on.
        :param str keyword: The keyword to match against.
        :return:            True if the function matches the keyword, False otherwise.
        :rtype:             bool
        """

        keyword = keyword.lower()

        if keyword in func.name.lower():
            return True
        demangled_name = func.demangled_name
        if demangled_name and keyword in demangled_name.lower():
            return True
        if isinstance(func.addr, int):
            if keyword in "%x" % func.addr:
                return True
            if keyword in "%#x" % func.addr:
                return True
        if keyword in ",".join(func.tags).lower():
            return True
        if func.binary and keyword in self._get_binary_name(func).lower():
            return True
        if extra_columns > 0:
            for idx in range(extra_columns):
                txt = self.workspace.plugins.extract_func_column(func, idx)[1]
                if txt and keyword in txt.lower():
                    return True
        return False


class QFunctionTableHeaderView(QHeaderView):
    """
    The header for QFunctionTableView.
    """

    def contextMenuEvent(self, event: "PySide6.QtGui.QContextMenuEvent") -> None:  # pylint:disable=unused-argument
        menu = QMenu("Column Menu", self)
        for idx in range(self.model().columnCount()):
            column_text = self.model().headerData(idx, Qt.Orientation.Horizontal, Qt.DisplayRole)
            action = QAction(column_text, self)
            action.setCheckable(True)
            hidden = self.isSectionHidden(idx)
            action.setChecked(not hidden)
            action.setEnabled(hidden or self.visibleSectionCount() > 1)
            action.toggled.connect(partial(self.setSectionVisible, idx))
            menu.addAction(action)
        menu.exec_(QCursor.pos())

    def visibleSectionCount(self):
        return self.model().columnCount() - self.hiddenSectionCount()

    def setSectionVisible(self, idx, visible):
        if visible or self.visibleSectionCount() > 1:
            self.setSectionHidden(idx, not visible)


class QFunctionTableView(QTableView):
    """
    The table view for QFunctionTable.
    """

    def __init__(self, parent, workspace, instance, selection_callback=None):
        super().__init__(parent)
        self.workspace = workspace
        self.instance = instance
        self._context_menu = FunctionContextMenu(workspace, self)

        self._function_table: QFunctionTable = parent
        self._selected_func = ObjectContainer(None, "Currently selected function")
        self._selected_func.am_subscribe(selection_callback)

        header = QFunctionTableHeaderView(Qt.Orientation.Horizontal, self)
        header.setSectionsClickable(True)
        self.setHorizontalHeader(header)
        self.horizontalHeader().setVisible(True)
        self.verticalHeader().setVisible(False)
        self.setSelectionBehavior(QAbstractItemView.SelectRows)
        self.setHorizontalScrollMode(QAbstractItemView.ScrollMode.ScrollPerPixel)

        self.show_alignment_functions = False
        self.filter_text = ""
        self._functions = None
        self._model = QFunctionTableModel(self.workspace, self.instance, [])

        self.setModel(self._model)

        self.horizontalHeader().setDefaultAlignment(Qt.AlignLeft)
        self.horizontalHeader().setSortIndicatorShown(True)
        self.horizontalHeader().setSectionResizeMode(QHeaderView.Interactive)
        self.horizontalHeader().setStretchLastSection(True)
        self.verticalHeader().setDefaultSectionSize(24)

        column_widths = [200, 80, 80, 80, 50, 50]
        for idx, width in enumerate(column_widths):
            self.setColumnWidth(idx, width)

        # slots
        self.horizontalHeader().sortIndicatorChanged.connect(self.sortByColumn)
        self.doubleClicked.connect(self._on_function_selected)

    def refresh(self, added_funcs: Optional[Set[int]] = None, removed_funcs: Optional[Set[int]] = None):
        if added_funcs:
            new_funcs = []
            for addr in added_funcs:
                try:
                    f_ = self._functions[addr]
                except KeyError:
                    continue
                if self.show_alignment_functions or (not self.show_alignment_functions and not f_.alignment):
                    new_funcs.append(f_)
            self._model.func_list += new_funcs
        if removed_funcs:
            self._model.func_list = [f_ for f_ in self._model.func_list if f_.addr not in removed_funcs]
        self.viewport().update()

    @property
    def function_manager(self):
        return self._functions

    @function_manager.setter
    def function_manager(self, functions):
        self._functions = functions
        self.load_functions()

    def subscribe_func_select(self, callback):
        self._selected_func.am_subscribe(callback)

    def filter(self, keyword):
        self.filter_text = keyword
        self._model.filter(keyword)

    def jump_to_result(self, index=0):
        if len(self._model.func_list) > index:
            self._selected_func.am_obj = self._model.func_list[index]
            self._selected_func.am_event(func=self._selected_func.am_obj)

    def load_functions(self):
        if not self.show_alignment_functions:
            self._model.func_list = [v for v in self._functions.values() if not v.alignment]
        else:
            self._model.func_list = list(self._functions.values())
        self._model.filter(self.filter_text)

    def _on_function_selected(self, model_index):
        row = model_index.row()
        self._selected_func.am_obj = self._model.func_list[row]
        self._selected_func.am_event(func=self._selected_func.am_obj)

    def keyPressEvent(self, key_event):
        text = key_event.text()
        if not text or text not in string.printable or text in string.whitespace:
            # modifier keys
            return super().keyPressEvent(key_event)

        # show the filtering text box
        self._function_table.show_filter_box(prefix=text)
        return True

    def contextMenuEvent(self, event: "PySide6.QtGui.QContextMenuEvent") -> None:  # pylint:disable=unused-argument
        rows = self.selectionModel().selectedRows()
        funcs = [self.instance.kb.functions[r.data()] for r in rows]
        self._context_menu.set(funcs).qmenu().popup(QCursor.pos())


class QFunctionTableFilterBox(QLineEdit):
    """
    The filter box for QFunctionTable.
    """

    def __init__(self, parent):
        super().__init__()

        self._table = parent

        self.installEventFilter(self)

    def eventFilter(self, obj, event):  # pylint:disable=unused-argument
        if event.type() == QEvent.KeyPress and event.key() == Qt.Key_Escape:
            if self.text():
                self.setText("")
            else:
                self._table.clear_filter_box()
            return True

        return False


class QFunctionTable(QWidget):
    """
    Implements a table for listing function details.
    """

    def __init__(self, parent, instance, selection_callback=None):
        super().__init__(parent)
        self.instance = instance

        self._view: FunctionsView = parent
        self._table_view: QFunctionTableView
        self._filter_box: QFunctionTableFilterBox
        self._toolbar: FunctionTableToolbar
        self._status_label: QLabel

        self._last_known_func_addrs: Set[int] = set()
        self._function_count = None

        self._init_widgets(selection_callback)

    @property
    def show_alignment_functions(self):
        return self._table_view.show_alignment_functions

    @property
    def function_manager(self) -> Optional["FunctionManager"]:
        if self._table_view is not None:
            return self._table_view.function_manager
        return None

    @function_manager.setter
    def function_manager(self, v):
        if v is not None:
            self._function_count = len(v)
        if self._table_view is not None:
            self._table_view.function_manager = v
        else:
            raise ValueError("QFunctionTableView is uninitialized.")
        self._last_known_func_addrs = {func.addr for func in self._table_view._model.func_list}
        self.filter_functions(self._filter_box.text())
        self.update_displayed_function_count()

    #
    # Public methods
    #

    def refresh(self):
        if self.function_manager is None:
            return

        if self._function_count != len(self.function_manager):
            # the number of functions has increased - we need to update the table
            added_funcs, removed_funcs = self._updated_functions(self.function_manager)
        else:
            added_funcs, removed_funcs = None, None

        self._table_view.refresh(added_funcs=added_funcs, removed_funcs=removed_funcs)
        self._function_count = len(self._last_known_func_addrs)
        self.update_displayed_function_count()

    def show_filter_box(self, prefix=""):
        if prefix:
            self._filter_box.setText(prefix)
        self._filter_box.show()
        self._filter_box.setFocus()

    def clear_filter_box(self):
        self._filter_box.setText("")
        self._table_view.setFocus()

    def toggle_show_alignment_functions(self):
        self._table_view.show_alignment_functions = not self._table_view.show_alignment_functions
        self._table_view.load_functions()
        self.update_displayed_function_count()

    def subscribe_func_select(self, callback):
        self._table_view.subscribe_func_select(callback)

    def update_displayed_function_count(self):
        cnt = self._table_view.model().rowCount()
        if self.function_manager is None:
            self._status_label.setText("")
            return
        if cnt == len(self.function_manager):
            self._status_label.setText("%d functions" % cnt)
        else:
            self._status_label.setText("%d/%d functions" % (cnt, len(self.function_manager)))

    def filter_functions(self, text):
        self._table_view.filter(text)
        self.update_displayed_function_count()

    #
    # Private methods
    #

    def _init_widgets(self, selection_callback=None):
        # function table view
        self._table_view = QFunctionTableView(self, self._view.workspace, self.instance, selection_callback)

        # filter text box
        self._filter_box = QFunctionTableFilterBox(self)
        self._filter_box.setClearButtonEnabled(True)
        self._filter_box.addAction(
            qta.icon("fa5s.search", color=Conf.palette_placeholdertext), QLineEdit.LeadingPosition
        )
        self._filter_box.setPlaceholderText("Filter by name...")
        self._filter_box.textChanged.connect(self._on_filter_box_text_changed)
        self._filter_box.returnPressed.connect(self._on_filter_box_return_pressed)

        # toolbar
        self._toolbar = FunctionTableToolbar(self)

        self._status_label = QLabel()

        status_lyt = QHBoxLayout()
        status_lyt.setContentsMargins(3, 3, 3, 3)
        status_lyt.setSpacing(3)
        status_lyt.addWidget(self._toolbar.qtoolbar())
        status_lyt.addWidget(self._filter_box)
        status_lyt.addWidget(self._status_label)

        # layout
        layout = QVBoxLayout()
        layout.addLayout(status_lyt)
        layout.addWidget(self._table_view)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(0)

        self.setLayout(layout)

    def _updated_functions(self, function_manager: "FunctionManager") -> Tuple[Set[int], Set[int]]:
        if len(self._last_known_func_addrs) == len(function_manager.function_addrs_set):
            return set(), set()
        new_func_addrs_set = function_manager.function_addrs_set.copy()
        added = new_func_addrs_set.difference(self._last_known_func_addrs)
        removed = self._last_known_func_addrs.difference(new_func_addrs_set)
        self._last_known_func_addrs = new_func_addrs_set
        return added, removed

    #
    # Events
    #

    def _on_filter_box_text_changed(self, text):
        self.filter_functions(text)

    def _on_filter_box_return_pressed(self):
        self._table_view.jump_to_result()
        self.clear_filter_box()



import logging
from pathlib import Path
from typing import Optional

from PySide6.QtGui import QColor
from PySide6.QtWidgets import QFileDialog

from angrmanagement.data.instance import Instance
from angrmanagement.data.jobs.cfg_generation import CFGGenerationJob
from angrmanagement.data.jobs.loading import LoadBinaryJob
from angrmanagement.plugins import BasePlugin
from angrmanagement.ui.views import DisassemblyView

from .diff_view import DiffDisassemblyView
from .function_diff import BFSFunctionDiff, FunctionDiff
from .settings_dialog import SettingsDialog

logger = logging.getLogger(__name__)


class PreciseDiffPlugin(BasePlugin):
    """
    A plugin for performing precisions diffs on two binaries loaded in angr-management.
    """

    DIFF_SETTINGS_CMD_EVT = 0
    LOAD_BINARY_CMD_EVT = 1
    RELOAD_BINARY_CMD_EVT = 2
    MENU_BUTTONS = ("Precise Diffing settings...", "Load binary for Precise Diffing...", "Refresh loaded diff binary")

    def __init__(self, workspace):
        super().__init__(workspace)
        self.diff_instance: Optional[Instance] = None
        self.current_revised_view: Optional[DisassemblyView] = None

        self.loaded_binary: Optional[Path] = None
        self.diff_algo: Optional[FunctionDiff] = None

        self.prefer_symbols = True
        self.resolve_strings = True
        self.use_addrs = False
        self.diff_algo_class = BFSFunctionDiff
        self.add_color = QColor(0xDDFFDD)
        self.del_color = QColor(0xFF7F7F)
        self.chg_color = QColor(0xF4ECC2)

        self.seen_insns = set()

    #
    # UI Callback Handlers
    #

    def handle_click_menu(self, idx):
        if idx == self.DIFF_SETTINGS_CMD_EVT:
            dialog = SettingsDialog(self)
            dialog.exec()
            if dialog.updates and self.loaded_binary:
                self.syncronize_with_original_disassembly_view()

        elif idx == self.LOAD_BINARY_CMD_EVT:
            filepath, _ = QFileDialog.getOpenFileName(caption="Load Recompiled Object")
            if not filepath:
                logger.warning("Binary selection cancelled.")
                return

            self.load_revised_binary_from_file(Path(filepath))

        elif idx == self.RELOAD_BINARY_CMD_EVT and self.loaded_binary is not None:
            self.load_revised_binary_from_file(self.loaded_binary)

    def color_insn(self, addr, selected, disasm_view):
        if disasm_view != self.current_revised_view:
            return None

        if self.diff_algo is None:
            return None

        diff_value = self.diff_algo.addr_diff_value(addr)
        return self._color_map(diff_value)

    #
    # View Construction
    #

    def _color_map(self, diff_value):
        diff_map = {
            FunctionDiff.OBJ_ADDED: self.add_color,
            FunctionDiff.OBJ_DELETED: self.del_color,
            FunctionDiff.OBJ_CHANGED: self.chg_color,
            FunctionDiff.OBJ_UNMODIFIED: None,
        }

        return diff_map[diff_value]

    def color_graph_diff(self, og_disasm: DisassemblyView, new_disasm: DisassemblyView):
        try:
            base_func = og_disasm.function.am_obj
            rev_func = new_disasm.function.am_obj
        except (AttributeError, ValueError):
            return

        if base_func is None or rev_func is None:
            return

        self.diff_algo = self.diff_algo_class(
            base_func,
            rev_func,
            disas_base=og_disasm.disasm,
            disas_rev=new_disasm.disasm,
            view_base=og_disasm,
            view_rev=new_disasm,
            resolve_strings=self.resolve_strings,
            prefer_symbols=self.prefer_symbols,
        )
        new_disasm.redraw_current_graph()

    def _destroy_recompiled_view(self):
        if self.current_revised_view:
            self.workspace.remove_view(self.current_revised_view)
            del self.current_revised_view

        if self.diff_instance:
            del self.diff_instance

    def _create_instance_from_binary(self, file_path: Path):
        self.diff_instance.recompilation_plugin = self
        self.diff_instance.workspace = self.workspace

        job = LoadBinaryJob(file_path, on_finish=self._create_instance_from_binary_done)
        self.loaded_binary = file_path
        self.diff_instance.add_job(job)

    def _create_instance_from_binary_done(self):
        job = CFGGenerationJob(on_finish=self._generate_binary_cfg_done)
        self.diff_instance.add_job(job)

    def _generate_binary_cfg_done(self):
        self.revised_binary_loaded()

    def _create_revised_disassembly_view(self):
        new_disass = DiffDisassemblyView(self.workspace, self.diff_instance, "center")
        new_disass.category = "diff"
        new_disass.base_caption = "Precise Diff"
        self.current_revised_view = new_disass
        self.workspace.add_view(self.current_revised_view)
        return self.current_revised_view

    def jump_to_in_revised_view(self, func):
        self.current_revised_view.display_function(func)
        self.current_revised_view.jump_to(func.addr)

    # pylint:disable=unused-argument
    def syncronize_with_original_disassembly_view(self, *args, **kwargs):
        og_view = self.workspace._get_or_create_view("disassembly", DisassemblyView)
        if not og_view:
            return

        try:
            func_obj = og_view.current_function
        except NotImplementedError:
            return

        if func_obj is None or func_obj.am_obj is None:
            return

        og_func = func_obj.am_obj
        og_func_name = og_func.name

        lookup_sym = og_func.addr if self.use_addrs else og_func_name
        try:
            revised_func = self.diff_instance.kb.functions[lookup_sym]
        except KeyError:
            logger.warning(
                "The function %s does not exist in the diffed binary",
                hex(lookup_sym) if self.use_addrs else og_func_name,
            )
            return

        self.jump_to_in_revised_view(revised_func)
        self.color_graph_diff(og_view, self.current_revised_view)

    def load_revised_binary_from_file(self, file_path: Path):
        self._destroy_recompiled_view()
        self.diff_instance = Instance()
        self._create_instance_from_binary(file_path)

    def revised_binary_loaded(self):
        self._create_revised_disassembly_view()
        self.syncronize_with_original_disassembly_view()
        self.workspace.view_manager.raise_view(self.current_revised_view)

        original_disass_view = self.diff_instance.workspace._get_or_create_view("disassembly", DisassemblyView)
        self.current_revised_view.sync_with_state_object(original_disass_view.sync_state)



import asyncio
import json
import logging
import math
import threading
import time

from PySide6.QtGui import QColor
from tornado.platform.asyncio import AnyThreadEventLoopPolicy

from angrmanagement.config import Conf
from angrmanagement.errors import UnexpectedStatusCodeError
from angrmanagement.logic.threads import gui_thread_schedule, gui_thread_schedule_async
from angrmanagement.plugins import BasePlugin
from angrmanagement.utils.io import download_url

from .parse_trace import trace_to_bb_addrs

log = logging.getLogger(__name__)

try:
    import slacrs.model
except ImportError:
    log.error("You don't have slacrs module installed locally, CoveragePlugin going to have a bad time.")


def generate_light_gradients(color, number, lightness=20):
    """
    return a List of QColors, where the colors are ordered in terms of
    lightness (last is given as color) the rest (total of number) are
    each lightness (%) lighter.
    """
    to_return = [color]
    for _ in range(number):
        last_color = to_return[-1]
        to_return.append(last_color.lighter(100 + lightness))
    to_return.reverse()
    return to_return


# TODO: This should really be a property of the target
TRACE_BASE = 0x4000000000


class CoveragePlugin(BasePlugin):
    """
    Implements the fuzzing coverage view.
    """

    def __init__(self, workspace):
        super().__init__(workspace)

        self.workspace = workspace

        # attempt to connect to the necessary components
        self.connector = self.workspace.plugins.get_plugin_instance_by_name("ChessConnector")

        if self.connector is None:
            self.workspace.log("Unable to retrieve plugin ChessConnector")

        self.slacrs_instance = self.connector.slacrs_instance()
        if self.slacrs_instance is None:
            self.workspace.log("Unable to retrieve Slacrs instance")

        self.dark_theme_color = QColor(0, 20, 147)
        self.light_theme_color = QColor(225, 174, 0)
        self.hit_color = self.dark_theme_color if Conf.theme_name == "dark" else self.light_theme_color
        self.num_gradients = 16
        self.gradients = generate_light_gradients(self.hit_color, self.num_gradients, lightness=int(100 / 16))

        self.running = False
        self.slacrs_thread = None

        self.seen_traces = None
        self.bbl_coverage = None
        self.bbl_coverage_hash = 0

        self.coverage_lock = threading.Lock()
        self.reset_coverage()

    MENU_BUTTONS = [
        "Start Showing Coverage",
        "Stop Showing Coverage",
    ]
    START_SHOWING_COVERAGE = 0
    STOP_SHOWING_COVERAGE = 1

    def handle_click_menu(self, idx):
        if idx < 0 or idx >= len(self.MENU_BUTTONS):
            return

        mapping = {
            self.START_SHOWING_COVERAGE: self.start,
            self.STOP_SHOWING_COVERAGE: self.stop,
        }

        mapping.get(idx)()

    def start(self):
        self.running = True
        self.slacrs_thread = threading.Thread(target=self.listen_for_events, daemon=True)
        self.slacrs_thread.start()
        gui_thread_schedule(self._refresh_gui)

    def stop(self):
        self.running = False
        gui_thread_schedule(self._refresh_gui)
        if self.workspace._main_window is not None:
            gui_thread_schedule_async(self.workspace._main_window.progress_done)

    def _coverage_of_func(self, func):
        """
        return (set of covered_bbls, and num_of_function_bbls)
        """
        func_bbls = func.block_addrs_set
        with self.coverage_lock:
            covered_bbls = self.bbl_coverage & func_bbls

        return covered_bbls, len(func_bbls)

    def color_block(self, addr):
        if not self.running:
            return None
        with self.coverage_lock:
            if addr in self.bbl_coverage:
                return self.dark_theme_color if Conf.theme_name == "dark" else self.light_theme_color
        return None

    def color_func(self, func):
        if not self.running:
            return None

        covered_bbls, total_bbls = self._coverage_of_func(func)

        # Be paranoid
        if total_bbls == 0:
            return None

        # Never want to highlight something that wasn't covered
        if len(covered_bbls) == 0:
            return None

        fraction_covered = len(covered_bbls) / total_bbls

        gradient_number = math.ceil(fraction_covered * len(self.gradients))
        return self.gradients[gradient_number - 1]

    FUNC_COLUMNS = ("Fuzzing Coverage",)

    def extract_func_column(self, func, idx):
        assert idx == 0
        if not self.running:
            return 0, "0%"

        covered_bbls, total_bbls = self._coverage_of_func(func)
        if len(covered_bbls) == 0:
            return 0, "0%"

        fraction_covered = len(covered_bbls) / total_bbls

        return fraction_covered, f"{int(round(fraction_covered*100,0))}%"

    def _refresh_gui(self):
        self.workspace.refresh()

    def reset_coverage(self):
        with self.coverage_lock:
            self.seen_traces = set()
            self.bbl_coverage = set()
            self.bbl_coverage_hash = 0

    def update_coverage_from_list(self, trace_addrs):
        log.info("Processing %d from the trace", len(trace_addrs))
        with self.coverage_lock:
            for addr in trace_addrs:
                self.bbl_coverage.add(addr)

        new_hash = hash(frozenset(self.bbl_coverage))
        if new_hash != self.bbl_coverage_hash:
            self.bbl_coverage_hash = new_hash
            gui_thread_schedule(self._refresh_gui)

    def update_coverage(self):
        self.set_status("Retrieving fuzzing coverage information...", 0.0)
        session = self.slacrs_instance.session()
        if session:
            for idx, trace in enumerate(
                session.query(slacrs.model.Trace)
                .filter(slacrs.model.Trace.input.has(target_image_id=self.connector.target_image_id))
                .order_by(slacrs.model.Trace.created_at)
            ):
                if not self.running:
                    break
                self.set_status(f"Processing trace {idx}...", 50.0)
                self.update_one_coverage(trace)
        self.set_status("Fuzzing coverage updated", 100.0)

    def update_one_coverage(self, trace):
        with self.coverage_lock:
            if trace.id in self.seen_traces:
                log.info("Already seen trace %s, skipping", trace.id)
                return

        log.info("Processing trace %s %s %s", trace.id, trace.input_id, trace.created_at)

        if not Conf.checrs_rest_endpoint_url:
            log.error("Unable to fetch trace %d because there is no CHECRS REST endpoint.", trace.id)
            return

        url = f"{Conf.checrs_rest_endpoint_url}v1/targets/{self.connector.target_image_id}/seeds/{trace.input_id}/trace"
        try:
            trace_bytes = download_url(url, parent=self.workspace._main_window, to_file=False)
        except UnexpectedStatusCodeError:
            log.exception("Unable to download %s.", url)
            return
        try:
            parsed_trace = json.loads(trace_bytes)
        except json.JSONDecodeError:
            log.exception("Unable to parse %s as JSON.", url)
            return

        bbl_addrs = trace_to_bb_addrs(parsed_trace, self.workspace.main_instance.project, TRACE_BASE)
        self.update_coverage_from_list(bbl_addrs)

        with self.coverage_lock:
            self.seen_traces.add(trace.id)
        log.info("Done processing trace %s.", trace.id)

    def listen_for_events(self):
        asyncio.set_event_loop_policy(AnyThreadEventLoopPolicy())
        while not self.connector and self.running:
            self.connector = self.workspace.plugins.get_plugin_instance_by_name("ChessConnector")
            time.sleep(1)

        while not self.slacrs_instance and self.running:
            self.slacrs_instance = self.connector.slacrs_instance()
            time.sleep(1)

        while not self.connector.target_image_id and self.running:
            time.sleep(1)

        if not self.running:
            return

        self.update_coverage()

        prev_target = self.connector.target_image_id
        while self.running:
            if self.connector.target_image_id != prev_target:
                self.reset_coverage()
                self.update_coverage()

            self.set_status("Retrieving fuzzing coverage information...", 0.0)
            new_event_count = self.slacrs_instance.fetch_events()
            trace_idx = 0
            for idx in range(new_event_count):
                e = self.slacrs_instance.event_queue.get_nowait()
                session = self.slacrs_instance.session()
                if e.kind == "trace":
                    obj = e.get_object(session)
                    if session.query(slacrs.model.Trace).filter_by(id=e.object_id) == 1:
                        if not self.running:
                            break
                        trace = session.query(slacrs.model.Trace).filter_by(obj.object_id).one()
                        self.set_status(f"Processing trace {trace_idx}...", idx * 100 / new_event_count)
                        trace_idx += 1
                        self.update_one_coverage(trace)
                session.close()
            self.set_status("Fuzzing coverage updated", 100.0)

    def set_status(self, status: str, percentage: float):
        if self.workspace._main_window is not None:
            gui_thread_schedule_async(self.workspace._main_window.progress, (status, percentage))



import logging

from PySide6.QtCore import QSize
from PySide6.QtWidgets import QHBoxLayout
from traitlets.config.configurable import MultipleInstanceError

from .view import BaseView

_l = logging.getLogger(name=__name__)


class ConsoleView(BaseView):
    """
    Console view providing IPython interactive session.
    """

    def __init__(self, workspace, instance, default_docking_position, *args, **kwargs):
        super().__init__("console", workspace, instance, default_docking_position, *args, **kwargs)

        self.base_caption = "Console"
        self._ipython_widget = None

        if self.workspace.main_window.initialized:
            self.mainWindowInitializedEvent()

    @property
    def ipython_widget_available(self):
        return self._ipython_widget is not None

    def mainWindowInitializedEvent(self):
        self._init_widgets()
        self.reload()

    def reload(self):
        if self._ipython_widget is None:
            return

        import angr  # pylint: disable=import-outside-toplevel,multiple-imports
        import claripy
        import cle

        namespace = {
            "angr": angr,
            "claripy": claripy,
            "cle": cle,
            "workspace": self.workspace,
            "instance": self.instance,
            "project": self.instance.project,
        }
        self._ipython_widget.push_namespace(namespace)

    def push_namespace(self, namespace):
        if self._ipython_widget is None:
            return

        self._ipython_widget.push_namespace(namespace)

    def print_text(self, msg):
        if self._ipython_widget is None:
            return

        self._ipython_widget.print_text(msg)

    def set_input_buffer(self, text):
        if self._ipython_widget is None:
            return
        self._ipython_widget.input_buffer = text

    @staticmethod
    def minimumSizeHint(*args, **kwargs):  # pylint: disable=unused-argument
        return QSize(0, 50)

    def _init_widgets(self):
        import angr  # pylint: disable=import-outside-toplevel,multiple-imports
        import claripy
        import cle

        namespace = {
            "angr": angr,
            "claripy": claripy,
            "cle": cle,
        }

        from angrmanagement.ui.widgets.qipython_widget import QIPythonWidget  # pylint:disable=import-outside-toplevel

        try:
            ipython_widget = QIPythonWidget(namespace=namespace)
        except MultipleInstanceError:
            _l.warning(
                "Fails to load the Console view since an IPython interpreter has already been loaded. "
                "You might be running angr Management with IPython."
            )
            return

        self._ipython_widget = ipython_widget
        ipython_widget.executed.connect(self.command_executed)

        hlayout = QHBoxLayout()
        hlayout.addWidget(ipython_widget)
        hlayout.setContentsMargins(0, 0, 0, 0)

        self.setLayout(hlayout)

    def command_executed(self, msg):
        if msg["msg_type"] == "execute_reply" and msg["content"]["status"] == "ok":
            view = self.workspace.view_manager.first_view_in_category("disassembly")
            if view is not None:
                view.refresh()

    def set_current_function(self, func):
        self.push_namespace(
            {
                "func": func,
                "function": func,
                "function_": func,  # deprecated
            }
        )



from PySide6.QtGui import QColor, QFont

COLOR_SCHEMES = {
    "Light": {
        "disasm_view_minimap_background_color": QColor(0xFF, 0xFF, 0xFF, 0xFF),
        "disasm_view_minimap_outline_color": QColor(0xB8, 0xB8, 0xB8, 0xFF),
        "disasm_view_operand_color": QColor(0x00, 0x00, 0x80),
        "disasm_view_operand_constant_color": QColor(0x10, 0x78, 0x96),
        "disasm_view_variable_label_color": QColor(0x00, 0x80, 0x00),
        "disasm_view_background_color": QColor(0xFF, 0xFF, 0xFF, 0xFF),
        "disasm_view_operand_highlight_color": QColor(0xFC, 0xEF, 0x00),
        "disasm_view_operand_select_color": QColor(0xFF, 0xFF, 0x00),
        "disasm_view_function_color": QColor(0x00, 0x00, 0xFF),
        "disasm_view_string_color": QColor(0xA0, 0xA0, 0xA4),
        "disasm_view_comment_color": QColor(0x37, 0x3D, 0x3F),
        "disasm_view_variable_ident_color": QColor(0xAA, 0x25, 0xDA),
        "disasm_view_variable_offset_color": QColor(0x80, 0x80, 0x00),
        "disasm_view_branch_target_text_color": QColor(0x80, 0x80, 0x00),
        "disasm_view_ir_default_color": QColor(0x80, 0x80, 0x80),
        "disasm_view_label_color": QColor(0x00, 0x00, 0xFF),
        "disasm_view_label_highlight_color": QColor(0xF0, 0xF0, 0xBF),
        "disasm_view_target_addr_color": QColor(0x00, 0x00, 0xFF),
        "disasm_view_antitarget_addr_color": QColor(0xFF, 0x00, 0x00),
        "disasm_view_node_shadow_color": QColor(0x00, 0x00, 0x00, 0x00),
        "disasm_view_node_background_color": QColor(0xFA, 0xFA, 0xFA),
        "disasm_view_node_zoomed_out_background_color": QColor(0xDA, 0xDA, 0xDA),
        "disasm_view_node_border_color": QColor(0xF0, 0xF0, 0xF0),
        "disasm_view_node_instruction_selected_background_color": QColor(0xB8, 0xC3, 0xD6),
        "disasm_view_node_address_color": QColor(0x00, 0x00, 0x00),
        "disasm_view_node_mnemonic_color": QColor(0x00, 0x00, 0x80),
        "disasm_view_selected_node_border_color": QColor(0x6B, 0x71, 0x7C),
        "disasm_view_printable_byte_color": QColor(0x00, 0x80, 0x40),
        "disasm_view_printable_character_color": QColor(0x00, 0x80, 0x40),
        "disasm_view_unprintable_byte_color": QColor(0x80, 0x40, 0x00),
        "disasm_view_unprintable_character_color": QColor(0x80, 0x40, 0x00),
        "disasm_view_unknown_byte_color": QColor(0xF0, 0x00, 0x00),
        "disasm_view_unknown_character_color": QColor(0xF0, 0x00, 0x00),
        "disasm_view_back_edge_color": QColor(0xF9, 0xD5, 0x77),
        "disasm_view_true_edge_color": QColor(0x79, 0xCC, 0xCD),
        "disasm_view_false_edge_color": QColor(0xF1, 0x66, 0x64),
        "disasm_view_direct_jump_edge_color": QColor(0x56, 0x5A, 0x5C),
        "disasm_view_exception_edge_color": QColor(0xF9, 0x91, 0x0A),
        "hex_view_selection_color": QColor(0xFF, 0x00, 0x00),
        "hex_view_selection_alt_color": QColor(0xA0, 0xA0, 0xA4),
        "hex_view_data_color": QColor(0x00, 0x00, 0xFF),
        "hex_view_string_color": QColor(0x00, 0xFF, 0xFF),
        "hex_view_instruction_color": QColor(0xFF, 0x00, 0xFF),
        "function_table_color": QColor(0x00, 0x00, 0x00),
        "function_table_syscall_color": QColor(0x00, 0x00, 0x80),
        "function_table_plt_color": QColor(0x00, 0x80, 0x00),
        "function_table_simprocedure_color": QColor(0x80, 0x00, 0x00),
        "function_table_alignment_color": QColor(0x80, 0x00, 0x80),
        "function_table_signature_bg_color": QColor(0xAA, 0xFF, 0xFF),
        "palette_window": QColor(0xEF, 0xEF, 0xEF, 0xFF),
        "palette_windowtext": QColor(0x00, 0x00, 0x00, 0xFF),
        "palette_base": QColor(0xFF, 0xFF, 0xFF, 0xFF),
        "palette_alternatebase": QColor(0xF7, 0xF7, 0xF7, 0xFF),
        "palette_tooltipbase": QColor(0xFF, 0xFF, 0xDC, 0xFF),
        "palette_tooltiptext": QColor(0x00, 0x00, 0x00, 0xFF),
        "palette_placeholdertext": QColor(0x00, 0x00, 0x00, 0xAF),
        "palette_text": QColor(0x00, 0x00, 0x00, 0xFF),
        "palette_button": QColor(0xEF, 0xEF, 0xEF, 0xFF),
        "palette_buttontext": QColor(0x00, 0x00, 0x00, 0xFF),
        "palette_brighttext": QColor(0xFF, 0xFF, 0xFF, 0xFF),
        "palette_highlight": QColor(0x30, 0x8C, 0xC6, 0xFF),
        "palette_highlightedtext": QColor(0xFF, 0xFF, 0xFF, 0xFF),
        "palette_disabled_text": QColor(0xBE, 0xBE, 0xBE, 0xFF),
        "palette_disabled_buttontext": QColor(0xBE, 0xBE, 0xBE, 0xFF),
        "palette_disabled_windowtext": QColor(0xBE, 0xBE, 0xBE, 0xFF),
        "palette_light": QColor(0xDF, 0xDF, 0xDF, 0xDF),
        "palette_midlight": QColor(0xCA, 0xCA, 0xCA, 0xFF),
        "palette_dark": QColor(0x9F, 0x9F, 0x9F, 0xFF),
        "palette_mid": QColor(0xB8, 0xB8, 0xB8, 0xFF),
        "palette_shadow": QColor(0x76, 0x76, 0x76, 0xFF),
        "palette_link": QColor(0x00, 0x00, 0xFF, 0xFF),
        "palette_linkvisited": QColor(0xFF, 0x00, 0xFF, 0xFF),
        "feature_map_regular_function_color": QColor(0x00, 0xA0, 0xE8),
        "feature_map_unknown_color": QColor(0x0A, 0x0A, 0x0A),
        "feature_map_delimiter_color": QColor(0x00, 0x00, 0x00),
        "feature_map_data_color": QColor(0xC0, 0xC0, 0xC0),
        "feature_map_string_color": QColor(0x00, 0xF0, 0x80),
        "pseudocode_comment_color": QColor(0x00, 0x80, 0x00, 0xFF),
        "pseudocode_function_color": QColor(0x00, 0x00, 0xFF, 0xFF),
        "pseudocode_library_function_color": QColor(0xFF, 0x00, 0xFF),
        "pseudocode_quotation_color": QColor(0x00, 0x80, 0x00, 0xFF),
        "pseudocode_keyword_color": QColor(0x00, 0x00, 0x80, 0xFF),
        "pseudocode_types_color": QColor(0x10, 0x78, 0x96),
        "pseudocode_variable_color": QColor(0x00, 0x00, 0x00, 0xFF),
        "pseudocode_label_color": QColor(0x00, 0x00, 0xFF),
        "pseudocode_highlight_color": QColor(0xFF, 0xFF, 0x00, 0xFF),
        "proximity_node_background_color": QColor(0xFA, 0xFA, 0xFA),
        "proximity_node_selected_background_color": QColor(0xCC, 0xCC, 0xCC),
        "proximity_node_border_color": QColor(0xF0, 0xF0, 0xF0),
        "proximity_function_node_text_color": QColor(0xFF, 0x00, 0x00),
        "proximity_string_node_text_color": QColor(0x00, 0x80, 0x00),
        "proximity_integer_node_text_color": QColor(0x00, 0x00, 0x80),
        "proximity_variable_node_text_color": QColor(0x00, 0x00, 0x80),
        "proximity_unknown_node_text_color": QColor(0x00, 0x00, 0x80),
        "proximity_call_node_text_color": QColor(0x00, 0x00, 0xFF),
        "proximity_call_node_text_color_plt": QColor(0x8B, 0x00, 0x8B),
        "proximity_call_node_text_color_simproc": QColor(0x8B, 0x00, 0x8B),
    },
    "Dark": {
        "disasm_view_minimap_background_color": QColor(0x28, 0x28, 0x28),
        "disasm_view_minimap_outline_color": QColor(0x40, 0x40, 0x40),
        "disasm_view_operand_color": QColor(0xF0, 0xF0, 0x5A),
        "disasm_view_operand_constant_color": QColor(0x34, 0xF0, 0x8C),
        "disasm_view_background_color": QColor(0x28, 0x28, 0x28),
        "disasm_view_variable_label_color": QColor(0x34, 0xD4, 0xF0),
        "disasm_view_operand_highlight_color": QColor(0x05, 0x2F, 0x50),
        "disasm_view_operand_select_color": QColor(0x09, 0x50, 0x8D),
        "disasm_view_function_color": QColor(0xC8, 0xC8, 0xC8),
        "disasm_view_string_color": QColor(0xA0, 0xA0, 0xA4),
        "disasm_view_comment_color": QColor(0xF5, 0xC2, 0x42, 0xBB),
        "disasm_view_variable_ident_color": QColor(0xF1, 0xA7, 0xFA),
        "disasm_view_variable_offset_color": QColor(0x80, 0x80, 0x00),
        "disasm_view_branch_target_text_color": QColor(0x80, 0x80, 0x00),
        "disasm_view_ir_default_color": QColor(0x80, 0x80, 0x80),
        "disasm_view_label_color": QColor(0x00, 0xAA, 0xFF),
        "disasm_view_label_highlight_color": QColor(0x2F, 0x2F, 0x25),
        "disasm_view_target_addr_color": QColor(0x00, 0xAA, 0xFF),
        "disasm_view_antitarget_addr_color": QColor(0xFF, 0x00, 0x00),
        "disasm_view_node_shadow_color": QColor(0x00, 0x00, 0x00, 0x4B),
        "disasm_view_node_background_color": QColor(0x3C, 0x3C, 0x3C),
        "disasm_view_node_zoomed_out_background_color": QColor(0x64, 0x64, 0x64),
        "disasm_view_node_border_color": QColor(0x50, 0x50, 0x50),
        "disasm_view_node_instruction_selected_background_color": QColor(0x4C, 0x50, 0x58),
        "disasm_view_node_address_color": QColor(0x2C, 0xC9, 0x76),
        "disasm_view_node_mnemonic_color": QColor(0xE0, 0xE0, 0xE0),
        "disasm_view_selected_node_border_color": QColor(0x6B, 0x71, 0x7C),
        "disasm_view_printable_byte_color": QColor(0x00, 0x80, 0x40),
        "disasm_view_printable_character_color": QColor(0x00, 0x80, 0x40),
        "disasm_view_unprintable_byte_color": QColor(0xBA, 0xBA, 0xBA),
        "disasm_view_unprintable_character_color": QColor(0xBA, 0xBA, 0xBA),
        "disasm_view_unknown_byte_color": QColor(0xF0, 0x00, 0x00),
        "disasm_view_unknown_character_color": QColor(0xF0, 0x00, 0x00),
        "disasm_view_back_edge_color": QColor(0xF9, 0xD5, 0x77),
        "disasm_view_true_edge_color": QColor(0x79, 0xCC, 0xCD),
        "disasm_view_false_edge_color": QColor(0xF1, 0x66, 0x64),
        "disasm_view_direct_jump_edge_color": QColor(0x56, 0x5A, 0x5C),
        "disasm_view_exception_edge_color": QColor(0xF9, 0x91, 0x0A),
        "hex_view_selection_color": QColor(0xFF, 0x00, 0x00),
        "hex_view_selection_alt_color": QColor(0xA0, 0xA0, 0xA4),
        "hex_view_data_color": QColor(0x00, 0x00, 0xFF),
        "hex_view_string_color": QColor(0x00, 0xFF, 0xFF),
        "hex_view_instruction_color": QColor(0xFF, 0x00, 0xFF),
        "function_table_color": QColor(0xE0, 0xE0, 0xE0),
        "function_table_syscall_color": QColor(0x00, 0x00, 0x80),
        "function_table_plt_color": QColor(0x00, 0x80, 0x00),
        "function_table_simprocedure_color": QColor(0x80, 0x00, 0x00),
        "function_table_alignment_color": QColor(0x80, 0x80, 0x00),
        "function_table_signature_bg_color": QColor(0xAA, 0xFF, 0xFF),
        "palette_window": QColor(0x35, 0x35, 0x35),
        "palette_windowtext": QColor(0xFF, 0xFF, 0xFF),
        "palette_base": QColor(0x28, 0x28, 0x28),
        "palette_alternatebase": QColor(0x1D, 0x1D, 0x1D),
        "palette_tooltipbase": QColor(0x35, 0x35, 0x35),
        "palette_tooltiptext": QColor(0xFF, 0xFF, 0xFF),
        "palette_placeholdertext": QColor(0xF8, 0xF8, 0xF8, 0xBB),
        "palette_text": QColor(0xE0, 0xE0, 0xE0),
        "palette_button": QColor(0x35, 0x35, 0x35),
        "palette_buttontext": QColor(0xE0, 0xE0, 0xE0),
        "palette_brighttext": QColor(0xFF, 0x00, 0x00),
        "palette_highlight": QColor(0x46, 0x85, 0xE5),
        "palette_highlightedtext": QColor(0xFF, 0xFF, 0xFF),
        "palette_disabled_text": QColor(0x80, 0x80, 0x80),
        "palette_disabled_buttontext": QColor(0x80, 0x80, 0x80),
        "palette_disabled_windowtext": QColor(0x80, 0x80, 0x80),
        "palette_light": QColor(0x46, 0x46, 0x46),
        "palette_midlight": QColor(0x40, 0x40, 0x40),
        "palette_dark": QColor(0x20, 0x20, 0x20),
        "palette_mid": QColor(0x28, 0x28, 0x28),
        "palette_shadow": QColor(0x16, 0x16, 0x16),
        "palette_link": QColor(0x2D, 0xC5, 0x2D).lighter(),
        "palette_linkvisited": QColor(0x2D, 0xC5, 0x2D).darker(),
        "feature_map_regular_function_color": QColor(0x00, 0xA0, 0xE8),
        "feature_map_unknown_color": QColor(0x0A, 0x0A, 0x0A),
        "feature_map_delimiter_color": QColor(0x00, 0x00, 0x00),
        "feature_map_data_color": QColor(0xC0, 0xC0, 0xC0),
        "feature_map_string_color": QColor(0x00, 0xF0, 0x80),
        "pseudocode_comment_color": QColor(0x00, 0x80, 0x00, 0xFF),
        "pseudocode_function_color": QColor(0x00, 0xAA, 0xFF),
        "pseudocode_library_function_color": QColor(0xAA, 0x00, 0xFF),
        "pseudocode_quotation_color": QColor(0x00, 0x80, 0x00, 0xFF),
        "pseudocode_keyword_color": QColor(0xF1, 0xA7, 0xFA),
        "pseudocode_types_color": QColor(0x00, 0xFF, 0xFF, 0xFF),
        "pseudocode_variable_color": QColor(0xE0, 0xE0, 0xE0),
        "pseudocode_label_color": QColor(0x00, 0xAA, 0xFF),
        "pseudocode_highlight_color": QColor(0x59, 0x44, 0x05, 0xFF),
        "proximity_node_background_color": QColor(0x3C, 0x3C, 0x3C),
        "proximity_node_selected_background_color": QColor(0x4C, 0x50, 0x58),
        "proximity_node_border_color": QColor(0x50, 0x50, 0x50),
        "proximity_function_node_text_color": QColor(0xFF, 0x00, 0xFF),
        "proximity_string_node_text_color": QColor(0xFA, 0xFA, 0xFA),
        "proximity_integer_node_text_color": QColor(0xFA, 0xFA, 0xFA),
        "proximity_variable_node_text_color": QColor(0xFA, 0xFA, 0xFA),
        "proximity_unknown_node_text_color": QColor(0xFA, 0xFA, 0xFA),
        "proximity_call_node_text_color": QColor(0x34, 0xD4, 0xF0),
        "proximity_call_node_text_color_plt": QColor(0x2C, 0xC9, 0x76),
        "proximity_call_node_text_color_simproc": QColor(0x2C, 0xC9, 0x76),
    },
    "Dracula": {
        "disasm_view_minimap_background_color": QColor(0x28, 0x2A, 0x36),
        "disasm_view_minimap_viewport_color": QColor(0x44, 0x47, 0x5A),
        "disasm_view_minimap_outline_color": QColor(0x44, 0x47, 0x5A),
        "disasm_view_operand_color": QColor(0xFF, 0x79, 0xC6),
        "disasm_view_operand_constant_color": QColor(0xF1, 0xFA, 0x8C),
        "disasm_view_background_color": QColor(0x1C, 0x1D, 0x26),
        "disasm_view_variable_label_color": QColor(0x34, 0xD4, 0xF0),
        "disasm_view_operand_highlight_color": QColor(0x05, 0x2F, 0x50),
        "disasm_view_operand_select_color": QColor(0x09, 0x50, 0x8D),
        "disasm_view_function_color": QColor(0xF8, 0xF8, 0xF2),
        "disasm_view_string_color": QColor(0xFF, 0xB8, 0x6C),
        "disasm_view_comment_color": QColor(0x5B, 0x65, 0x8E),
        "disasm_view_variable_ident_color": QColor(0xBD, 0x93, 0xF9),
        "disasm_view_ir_default_color": QColor(0x50, 0xFA, 0x7B),
        "disasm_view_label_color": QColor(0x00, 0xAA, 0xFF),
        "disasm_view_label_highlight_color": QColor(0x2F, 0x2F, 0x25),
        "disasm_view_target_addr_color": QColor(0x50, 0xFA, 0x7B),
        "disasm_view_antitarget_addr_color": QColor(0xFF, 0x6E, 0x6E),
        "disasm_view_node_shadow_color": QColor(0x00, 0x00, 0x00),
        "disasm_view_node_background_color": QColor(0x28, 0x2A, 0x36),
        "disasm_view_node_zoomed_out_background_color": QColor(0x5D, 0x62, 0x7E),
        "disasm_view_node_border_color": QColor(0x50, 0x50, 0x50),
        "disasm_view_node_instruction_selected_background_color": QColor(0x44, 0x47, 0x5A),
        "disasm_view_node_address_color": QColor(0x3F, 0xA1, 0x5D),
        "disasm_view_node_mnemonic_color": QColor(0xF8, 0xF8, 0xF2),
        "disasm_view_selected_node_border_color": QColor(0x6B, 0x71, 0x7C),
        "disasm_view_printable_byte_color": QColor(0x50, 0xFA, 0x7B),
        "disasm_view_printable_character_color": QColor(0x50, 0xFA, 0x7B),
        "disasm_view_unprintable_byte_color": QColor(0xBA, 0xBA, 0xBA),
        "disasm_view_unprintable_character_color": QColor(0xBA, 0xBA, 0xBA),
        "disasm_view_unknown_byte_color": QColor(0xFF, 0x6E, 0x6E),
        "disasm_view_unknown_character_color": QColor(0xFF, 0x6E, 0x6E),
        "disasm_view_back_edge_color": QColor(0x9B, 0xA1, 0x67),
        "disasm_view_true_edge_color": QColor(0x3F, 0xA1, 0x5D),
        "disasm_view_false_edge_color": QColor(0xA3, 0x43, 0x48),
        "disasm_view_direct_jump_edge_color": QColor(0x56, 0x5A, 0x5C),
        "disasm_view_exception_edge_color": QColor(0xFF, 0xB8, 0x6C),
        "hex_view_selection_color": QColor(0x44, 0x47, 0x5A),
        "hex_view_selection_alt_color": QColor(0x9B, 0x9E, 0xB3),
        "hex_view_data_color": QColor(0x61, 0x97, 0xA8),
        "hex_view_string_color": QColor(0x9B, 0xA1, 0x67),
        "hex_view_instruction_color": QColor(0xA3, 0x57, 0x88),
        "function_table_color": QColor(0xF8, 0xF8, 0xF2),
        "function_table_syscall_color": QColor(0x00, 0x00, 0x80),
        "function_table_plt_color": QColor(0x50, 0xFA, 0x7B),
        "function_table_simprocedure_color": QColor(0xFF, 0x55, 0x55),
        "function_table_alignment_color": QColor(0xF1, 0xFA, 0x8C),
        "function_table_signature_bg_color": QColor(0x8B, 0xE9, 0xFD),
        "palette_window": QColor(0x24, 0x26, 0x31),
        "palette_windowtext": QColor(0xF8, 0xF8, 0xF2),
        "palette_base": QColor(0x28, 0x2A, 0x36),
        "palette_alternatebase": QColor(0x3A, 0x3D, 0x4E),
        "palette_tooltipbase": QColor(0x24, 0x26, 0x31),
        "palette_tooltiptext": QColor(0xF8, 0xF8, 0xF2),
        "palette_placeholdertext": QColor(0xF8, 0xF8, 0xF2, 0xBB),
        "palette_text": QColor(0xF8, 0xF8, 0xF2),
        "palette_button": QColor(0x24, 0x26, 0x31),
        "palette_buttontext": QColor(0xF8, 0xF8, 0xF2),
        "palette_brighttext": QColor(0xFF, 0xFF, 0xFF),
        "palette_highlight": QColor(0x44, 0x47, 0x5A),
        "palette_highlightedtext": QColor(0xF8, 0xF8, 0xF2),
        "palette_disabled_text": QColor(0x80, 0x80, 0x80),
        "palette_disabled_buttontext": QColor(0x80, 0x80, 0x80),
        "palette_disabled_windowtext": QColor(0x80, 0x80, 0x80),
        "palette_light": QColor(0x41, 0x44, 0x57),
        "palette_midlight": QColor(0x3B, 0x3E, 0x50),
        "palette_dark": QColor(0x1C, 0x1D, 0x26),
        "palette_mid": QColor(0x28, 0x2A, 0x36),
        "palette_shadow": QColor(0x16, 0x17, 0x1E),
        "palette_link": QColor(0x8B, 0xE9, 0xFD),
        "palette_linkvisited": QColor(0xBD, 0x93, 0xF9),
        "pseudocode_comment_color": QColor(0x62, 0x72, 0xA4),
        "pseudocode_comment_weight": QFont.Weight.Normal,
        "pseudocode_function_color": QColor(0x50, 0xFA, 0x7B),
        "pseudocode_library_function_color": QColor(0x8B, 0xE9, 0xFD),
        "pseudocode_quotation_color": QColor(0xF1, 0xFA, 0x8C),
        "pseudocode_keyword_color": QColor(0xFF, 0x79, 0xC6),
        "pseudocode_types_color": QColor(0x8B, 0xE9, 0xFD),
        "pseudocode_types_style": QFont.Style.StyleItalic,
        "pseudocode_variable_color": QColor(0xF8, 0xF8, 0xF2),
        "pseudocode_label_color": QColor(0x00, 0xAA, 0xFF),
        "pseudocode_highlight_color": QColor(0x44, 0x47, 0x5A),
        "proximity_node_background_color": QColor(0x28, 0x2A, 0x36),
        "proximity_node_selected_background_color": QColor(0x4C, 0x50, 0x58),
        "proximity_node_border_color": QColor(0x50, 0x50, 0x50),
        "proximity_function_node_text_color": QColor(0xF8, 0xF8, 0xF2),
        "proximity_string_node_text_color": QColor(0xF1, 0xFA, 0x8C),
        "proximity_integer_node_text_color": QColor(0xF1, 0xFA, 0x8C),
        "proximity_variable_node_text_color": QColor(0xFF, 0x79, 0xC6),
        "proximity_unknown_node_text_color": QColor(0xF8, 0xF8, 0xF2),
        "proximity_call_node_text_color": QColor(0x34, 0xD4, 0xF0),
        "proximity_call_node_text_color_plt": QColor(0x50, 0xFA, 0x7B),
        "proximity_call_node_text_color_simproc": QColor(0xFF, 0x55, 0x55),
        "feature_map_regular_function_color": QColor(0x50, 0xFA, 0x7B),
        "feature_map_unknown_color": QColor(0x28, 0x2A, 0x36),
        "feature_map_delimiter_color": QColor(0x00, 0x00, 0x00),
        "feature_map_data_color": QColor(0x5D, 0x62, 0x7E),
        "feature_map_string_color": QColor(0xF1, 0xFA, 0x8C),
    },
    "Catppuccin Mocha": {
        # Disassembly View Colors
        "disasm_view_minimap_background_color": QColor(30, 30, 46),  # base
        "disasm_view_minimap_outline_color": QColor(88, 91, 112),  # surface2
        "disasm_view_operand_color": QColor(137, 220, 235),  # sky
        "disasm_view_operand_constant_color": QColor(249, 226, 175),  # yellow
        "disasm_view_variable_label_color": QColor(166, 227, 161),  # green
        "disasm_view_background_color": QColor(30, 30, 46),  # base
        "disasm_view_function_color": QColor(166, 227, 161),  # green
        "disasm_view_string_color": QColor(249, 226, 175),  # yellow
        "disasm_view_comment_color": QColor(205, 214, 244),  # text
        "disasm_view_variable_ident_color": QColor(180, 190, 254),  # lavender
        "disasm_view_variable_offset_color": QColor(243, 139, 168),  # red
        "disasm_view_branch_target_text_color": QColor(137, 180, 250),  # blue
        "disasm_view_ir_default_color": QColor(88, 91, 112),  # surface2
        "disasm_view_label_color": QColor(137, 220, 235),  # sky
        "disasm_view_label_highlight_color": QColor(105, 110, 150),
        "disasm_view_target_addr_color": QColor(166, 227, 161),  # green
        "disasm_view_antitarget_addr_color": QColor(243, 139, 168),  # red
        "disasm_view_node_shadow_color": QColor(0, 0, 0, 0),  # transparent
        "disasm_view_node_background_color": QColor(30, 30, 46),  # base
        "disasm_view_node_zoomed_out_background_color": QColor(49, 50, 68),  # surface0
        "disasm_view_node_border_color": QColor(49, 50, 68),  # surface0
        "disasm_view_node_address_color": QColor(205, 214, 244),  # text
        "disasm_view_node_mnemonic_color": QColor(137, 220, 235),  # sky
        "disasm_view_printable_byte_color": QColor(166, 227, 161),  # green
        "disasm_view_printable_character_color": QColor(166, 227, 161),  # green
        "disasm_view_unprintable_byte_color": QColor(243, 139, 168),  # red
        "disasm_view_unprintable_character_color": QColor(243, 139, 168),  # red
        "disasm_view_unknown_byte_color": QColor(137, 220, 235),  # sky
        "disasm_view_unknown_character_color": QColor(137, 220, 235),  # sky
        "disasm_view_back_edge_color": QColor(166, 227, 161),  # green
        "disasm_view_true_edge_color": QColor(166, 227, 161),  # green
        "disasm_view_false_edge_color": QColor(243, 139, 168),  # red
        "disasm_view_direct_jump_edge_color": QColor(137, 220, 235),  # sky
        "disasm_view_exception_edge_color": QColor(243, 139, 168),  # red
        # Hex View Colors
        "hex_view_selection_color": QColor(105, 110, 150),  # Custom highlight
        "hex_view_selection_alt_color": QColor(105, 110, 150),  # Custom highlight
        "hex_view_data_color": QColor(137, 180, 250),  # blue
        "hex_view_string_color": QColor(249, 226, 175),  # yellow
        "hex_view_instruction_color": QColor(166, 227, 161),  # green
        # Function Table Colors
        "function_table_color": QColor(205, 214, 244),  # text
        "function_table_syscall_color": QColor(243, 139, 168),  # red
        "function_table_plt_color": QColor(166, 227, 161),  # green
        "function_table_simprocedure_color": QColor(180, 190, 254),  # lavender
        "function_table_alignment_color": QColor(249, 226, 175),  # yellow
        "function_table_signature_bg_color": QColor(245, 224, 220),  # rosewater
        # Palette Colors
        "palette_window": QColor(30, 30, 46),  # base
        "palette_windowtext": QColor(205, 214, 244),  # text
        "palette_base": QColor(30, 30, 46),  # base
        "palette_alternatebase": QColor(49, 50, 68),  # surface0
        "palette_tooltipbase": QColor(30, 30, 46),  # base
        "palette_tooltiptext": QColor(88, 91, 112),  # surface2
        "palette_text": QColor(205, 214, 244),  # text
        "palette_button": QColor(30, 30, 46),  # base
        "palette_buttontext": QColor(88, 91, 112),  # surface2
        "palette_brighttext": QColor(88, 91, 112),  # surface2
        "palette_highlight": QColor(105, 110, 150),  # Custom highlight
        "palette_highlightedtext": QColor(255, 255, 255),  # White text for contrast
        # Feature Map Colors
        "feature_map_regular_function_color": QColor(166, 227, 161),  # green
        "feature_map_unknown_color": QColor(137, 220, 235),  # sky
        "feature_map_delimiter_color": QColor(249, 226, 175),  # yellow
        "feature_map_data_color": QColor(180, 190, 254),  # lavender
        "feature_map_string_color": QColor(249, 226, 175),  # yellow
        # Pseudocode Colors
        "pseudocode_comment_color": QColor(205, 214, 244),  # text
        "pseudocode_function_color": QColor(166, 227, 161),  # green
        "pseudocode_library_function_color": QColor(243, 139, 168),  # red
        "pseudocode_quotation_color": QColor(166, 227, 161),  # green
        "pseudocode_keyword_color": QColor(137, 220, 235),  # sky
        "pseudocode_types_color": QColor(249, 226, 175),  # yellow
        "pseudocode_variable_color": QColor(205, 214, 244),  # text
        "pseudocode_label_color": QColor(137, 220, 235),  # sky
        "pseudocode_highlight_color": QColor(105, 110, 150),  # Custom highlight
        # Proximity View Colors
        "proximity_node_background_color": QColor(30, 30, 46),  # base
        "proximity_node_selected_background_color": QColor(105, 110, 150),  # Custom highlight
        "proximity_node_border_color": QColor(88, 91, 112),  # surface2
        "proximity_function_node_text_color": QColor(166, 227, 161),  # green
        "proximity_string_node_text_color": QColor(249, 226, 175),  # yellow
        "proximity_integer_node_text_color": QColor(137, 220, 235),  # sky
        "proximity_variable_node_text_color": QColor(180, 190, 254),  # lavender
        "proximity_unknown_node_text_color": QColor(245, 224, 220),  # rosewater
        "proximity_call_node_text_color": QColor(166, 227, 161),  # green
        "proximity_call_node_text_color_plt": QColor(137, 180, 250),  # blue
        "proximity_call_node_text_color_simproc": QColor(243, 139, 168),  # red
        # Palette Colors
        "palette_mid": QColor(68, 71, 90),  # A mid-tone color between surface0 and base
        "palette_placeholdertext": QColor(137, 220, 235, 128),  # Sky color with 50% opacity for placeholder text
        "palette_light": QColor(68, 71, 90),  # Lighter than base but not too bright
        "palette_midlight": QColor(58, 60, 78),  # A slightly lighter surface color for a subtle lift
        "palette_dark": QColor(20, 20, 30),  # A darker version of the base for depth
        "palette_shadow": QColor(0, 0, 0, 50),  # A soft shadow with 50% opacity
        "palette_link": QColor(137, 220, 235),  # Sky color for links
        "palette_linkvisited": QColor(180, 190, 254),  # Lavender color to indicate visited links
        # Disabled Colors
        "palette_disabled_text": QColor(88, 91, 112, 128),  # Surface2 color with 50% opacity for disabled text
        "palette_disabled_buttontext": QColor(
            88, 91, 112, 128
        ),  # Surface2 color with 50% opacity for disabled button text
        "palette_disabled_windowtext": QColor(
            88, 91, 112, 128
        ),  # Surface2 color with 50% opacity for disabled window text
        # Disassembly View Colors
        "disasm_view_node_instruction_selected_background_color": QColor(
            105, 110, 150
        ),  # Custom highlight for selected instruction background
        "disasm_view_operand_highlight_color": QColor(
            105, 110, 150, 128
        ),  # Custom highlight with 50% opacity for operand highlight
        "disasm_view_selected_node_border_color": QColor(105, 110, 150),  # Custom highlight for selected node border
        "disasm_view_operand_select_color": QColor(105, 110, 150),  # Custom highlight for operand selection
    },
}



# pylint:disable=no-self-use
import datetime
import logging
import os
import pickle
import time
from functools import partial
from typing import TYPE_CHECKING, Optional

import angr
import angr.flirt
import PySide6QtAds as QtAds
import qtawesome as qta
from angr.angrdb import AngrDB
from PySide6.QtCore import QEvent, QObject, QSize, Qt, QTimer, QUrl
from PySide6.QtGui import QDesktopServices, QIcon, QKeySequence, QShortcut, QWindow
from PySide6.QtWidgets import (
    QFileDialog,
    QHBoxLayout,
    QLabel,
    QMainWindow,
    QMessageBox,
    QProgressBar,
    QProgressDialog,
    QWidget,
)

from angrmanagement.config import IMG_LOCATION, Conf, save_config
from angrmanagement.daemon import daemon_conn, daemon_exists, run_daemon_process
from angrmanagement.daemon.client import ClientService
from angrmanagement.data.instance import Instance
from angrmanagement.data.jobs import DependencyAnalysisJob
from angrmanagement.data.jobs.loading import LoadAngrDBJob, LoadBinaryJob, LoadTargetJob
from angrmanagement.data.library_docs import LibraryDocs
from angrmanagement.errors import InvalidURLError, UnexpectedStatusCodeError
from angrmanagement.logic import GlobalInfo
from angrmanagement.logic.commands import BasicCommand
from angrmanagement.ui.views import DisassemblyView
from angrmanagement.utils.env import app_root, is_pyinstaller
from angrmanagement.utils.io import download_url, isurl

from .dialogs.about import LoadAboutDialog
from .dialogs.command_palette import CommandPaletteDialog, GotoPaletteDialog
from .dialogs.load_docker_prompt import LoadDockerPrompt, LoadDockerPromptError
from .dialogs.load_plugins import LoadPlugins
from .dialogs.new_state import NewState
from .dialogs.preferences import Preferences
from .menus.analyze_menu import AnalyzeMenu
from .menus.file_menu import FileMenu
from .menus.help_menu import HelpMenu
from .menus.plugin_menu import PluginMenu
from .menus.view_menu import ViewMenu
from .toolbar_manager import ToolbarManager
from .toolbars import DebugToolbar, FeatureMapToolbar, FileToolbar
from .widgets import QIconLabel
from .workspace import Workspace

try:
    import archr
    import keystone
except ImportError:
    archr = None
    keystone = None

if TYPE_CHECKING:
    from PySide6.QtWidgets import QApplication


_l = logging.getLogger(name=__name__)


class DockShortcutEventFilter(QObject):
    """
    Filter to support shortcuts on floating dock windows that overlap main window registered menu action shortcuts.
    """

    def __init__(self, main_window: "MainWindow"):
        super().__init__()
        self._main_window: MainWindow = main_window

    def eventFilter(self, qobject, event):
        if event.type() == QEvent.KeyPress and QKeySequence(event.keyCombination()) == QKeySequence("Ctrl+Shift+P"):
            self._main_window.show_command_palette(qobject)
            return True
        return False


class ShiftShiftEventFilter(QObject):
    """
    Filter to catch Shift+Shift key sequence for goto-anything activation.
    """

    activation_key = Qt.Key.Key_Shift
    activation_count: int = 2
    timeout_secs: float = 1

    def __init__(self, main_window: "MainWindow"):
        super().__init__()
        self._main_window: MainWindow = main_window
        self._press_count: int = 0
        self._last_press_time: float = 0
        self._did_process_qwindow_event: bool = False

    def eventFilter(self, qobject, event):
        # Key Event propagation will begin at QWindow and continue down the widget tree. Use KeyEvent on QWindow to
        # distinguish unique key presses, then intercept KeyEvent at first QWidget the event is propagated to.

        if event.type() == QEvent.KeyPress:
            if isinstance(qobject, QWindow) and qobject.modality() == Qt.WindowModality.NonModal:
                self._did_process_qwindow_event = True
                return False
            if not isinstance(qobject, QWidget) or not self._did_process_qwindow_event:
                return False
            self._did_process_qwindow_event = False

            if event.count() == 1 and event.key() == self.activation_key:
                now = time.time()
                if now - self._last_press_time >= self.timeout_secs:
                    self._press_count = 0
                self._last_press_time = now

                self._press_count += 1
                if self._press_count >= self.activation_count:
                    self._press_count = 0
                    self._main_window.show_goto_palette(qobject)
                    return True

            else:
                self._press_count = 0

        return False


class MainWindow(QMainWindow):
    """
    The main window of angr management.
    """

    def __init__(self, app: Optional["QApplication"] = None, parent=None, show=True, use_daemon=False):
        super().__init__(parent)
        self.initialized = False

        icon_location = os.path.join(IMG_LOCATION, "angr.png")
        self.setWindowIcon(QIcon(icon_location))
        self.setWindowTitle("angr management")

        GlobalInfo.main_window = self
        self.shown_at_start = show

        # initialization
        self.setMinimumSize(QSize(400, 400))

        self.app: Optional[QApplication] = app
        self.workspace: Workspace = None
        self.dock_manager: QtAds.CDockManager
        self._dock_shortcut_event_filter = DockShortcutEventFilter(self)

        self._shift_shift_event_filter = ShiftShiftEventFilter(self)
        if app:
            self.app.installEventFilter(self._shift_shift_event_filter)

        self.toolbar_manager: ToolbarManager = ToolbarManager(self)

        self._progress_stopwatch_start_time: float = 0.0
        self._progress_message: str = ""
        self._progress_percentage: float = 0
        self._progress_update_timer: QTimer = QTimer()
        self._progress_update_timer.setSingleShot(False)
        self._progress_update_timer.setInterval(1000)
        self._progress_update_timer.timeout.connect(self._on_progress_update_timer_timeout)

        self.defaultWindowFlags = None

        # menus
        self._file_menu = None  # FileMenu
        self._analyze_menu = None
        self._view_menu = None
        self._help_menu = None
        self._plugin_menu = None

        self._init_statusbar()
        self._init_workspace()
        self._init_toolbars()
        self._init_menus()
        self._init_plugins()
        self._init_library_docs()
        # self._init_url_scheme_handler()

        self._register_commands()

        self.workspace.plugins.on_workspace_initialized(self)

        self._init_shortcuts()
        self._init_flirt_signatures()

        self._run_daemon(use_daemon=use_daemon)

        # I'm ready to show off!
        if show:
            self.showMaximized()
            self.windowHandle().screenChanged.connect(self.on_screen_changed)
            self.show()

    def sizeHint(self, *args, **kwargs):  # pylint: disable=unused-argument,no-self-use
        return QSize(1200, 800)

    #
    # Properties
    #

    @property
    def caption(self):
        return self.getWindowTitle()

    @caption.setter
    def caption(self, v):
        self.setWindowTitle(v)

    #
    # Dialogs
    #

    def open_mainfile_dialog(self):
        # pylint: disable=assigning-non-slot
        # https://github.com/PyCQA/pylint/issues/3793
        file_path, _ = QFileDialog.getOpenFileName(
            self,
            "Open a binary",
            self._get_recent_dir(),
            "All executables (*);;Windows PE files (*.exe);;Core Dumps (*.core);;angr database (*.adb)",
        )
        return file_path

    def _pick_image_dialog(self):
        try:
            prompt = LoadDockerPrompt(parent=self)
        except LoadDockerPromptError:
            return None
        if prompt.exec_() == 0:
            return None  # User canceled
        return prompt.textValue()

    def open_load_plugins_dialog(self):
        dlg = LoadPlugins(self.workspace.plugins)
        dlg.setModal(True)
        dlg.exec_()

    def open_newstate_dialog(self):
        if self.workspace.main_instance.project.am_none:
            QMessageBox.critical(self, "Cannot create new states", "Please open a binary to analyze first.")
            return
        new_state_dialog = NewState(self.workspace, self.workspace.main_instance, parent=self, create_simgr=True)
        new_state_dialog.exec_()

    def open_doc_link(self):
        QDesktopServices.openUrl(QUrl("https://docs.angr.io/", QUrl.TolerantMode))

    def open_about_dialog(self):
        dlg = LoadAboutDialog()
        dlg.exec_()

    #
    # Widgets
    #

    def _init_statusbar(self):
        layout = QHBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addStretch()

        self._status_label = QLabel()
        self._status_label.hide()
        layout.addWidget(self._status_label)

        self._stopwatch_label = QIconLabel(qta.icon("fa5s.stopwatch", color=Conf.palette_buttontext))
        self._stopwatch_label.hide()
        layout.addWidget(self._stopwatch_label)

        self._interrupt_job_button = QIconLabel(qta.icon("fa5s.times-circle", color=Conf.palette_buttontext))
        self._interrupt_job_button.clicked.connect(self.interrupt_current_job)
        self._interrupt_job_button.hide()
        layout.addWidget(self._interrupt_job_button)

        self._progress_label = QLabel()
        self._progress_label.hide()
        layout.addWidget(self._progress_label)

        self._progress_bar = QProgressBar()
        self._progress_bar.setMinimum(0)
        self._progress_bar.setMaximum(100)
        self._progress_bar.setMinimumWidth(100)
        self._progress_bar.setMaximumWidth(100)
        self._progress_bar.hide()
        layout.addWidget(self._progress_bar)

        container_widget = QWidget()
        container_widget.setLayout(layout)
        self.statusBar().addPermanentWidget(container_widget)

        self._progress_dialog = QProgressDialog("Waiting...", "Cancel", 0, 100, self)
        self._progress_dialog.setAutoClose(False)
        self._progress_dialog.setWindowFlags(self._progress_dialog.windowFlags() & ~Qt.WindowContextHelpButtonHint)
        self._progress_dialog.setModal(True)
        self._progress_dialog.setMinimumDuration(2**31 - 1)

        def on_cancel():
            if self.workspace is None:
                return
            for job in self.workspace.main_instance.jobs:
                if job.blocking:
                    job.keyboard_interrupt()
                    break

        self._progress_dialog.canceled.connect(on_cancel)
        self._progress_dialog.close()

    def _init_toolbars(self):
        for cls in (FileToolbar, DebugToolbar, FeatureMapToolbar):
            self.toolbar_manager.show_toolbar_by_class(cls)

    #
    # Menus
    #

    def _init_menus(self):
        self._file_menu = FileMenu(self)
        self._analyze_menu = AnalyzeMenu(self)
        self._view_menu = ViewMenu(self)
        self._help_menu = HelpMenu(self)
        self._plugin_menu = PluginMenu(self)

        for path in Conf.recent_files:
            self._file_menu.add_recent(path)

        self.menuBar().addMenu(self._file_menu.qmenu())
        self.menuBar().addMenu(self._view_menu.qmenu())
        self.menuBar().addMenu(self._analyze_menu.qmenu())
        self.menuBar().addMenu(self._plugin_menu.qmenu())
        self.menuBar().addMenu(self._help_menu.qmenu())

    #
    # Workspace
    #

    def _init_workspace(self):
        """
        Initialize workspace

        :return:    None
        """
        QtAds.CDockManager.setConfigFlags(
            (
                QtAds.CDockManager.DefaultBaseConfig
                | QtAds.CDockManager.OpaqueSplitterResize
                | QtAds.CDockManager.FocusHighlighting
            )
            & ~QtAds.CDockManager.DockAreaHasUndockButton
        )
        self.dock_manager = QtAds.CDockManager(self)
        self.dock_manager.setStyleSheet("")  # Clear stylesheet overrides
        self.setCentralWidget(self.dock_manager)
        wk = Workspace(self, Instance())
        self.workspace = wk

        def set_caption(**kwargs):  # pylint: disable=unused-argument
            if self.workspace.main_instance.project.am_none:
                self.caption = ""
            elif self.workspace.main_instance.project.filename is None:
                self.caption = "Loaded from stream"
            else:
                self.caption = os.path.basename(self.workspace.main_instance.project.filename)

        self.workspace.main_instance.project.am_subscribe(set_caption)

    #
    # Shortcuts
    #

    def interrupt_current_job(self):
        self.workspace.main_instance.interrupt_current_job()

    def _init_shortcuts(self):
        """
        Initialize shortcuts

        :return:    None
        """

        for i in range(1, 10):
            QShortcut(QKeySequence(f"Alt+{i}"), self, lambda idx=i: self._raise_view(idx - 1))
        QShortcut(QKeySequence("Alt+0"), self, lambda: self._raise_view(9))

        QShortcut(QKeySequence("Ctrl+I"), self, self.interrupt_current_job)

        # Raise the DisassemblyView after everything has initialized
        self._raise_view(0)

        # Toggle exec breakpoint
        QShortcut(QKeySequence(Qt.Key_F2), self, self.workspace.toggle_exec_breakpoint)

        # Single step
        QShortcut(QKeySequence(Qt.Key_F7), self, self.workspace.step_forward)

        # Run
        QShortcut(QKeySequence(Qt.Key_F9), self, self.workspace.continue_forward)

    def init_shortcuts_on_dock(self, dock_widget):
        """
        Installs an event filter on the dock widget to support floating dock global shortcuts (e.g. command palette).
        """
        dock_widget.installEventFilter(self._dock_shortcut_event_filter)

    #
    # Plugins
    #

    def _init_plugins(self):
        self.workspace.plugins.discover_and_initialize_plugins()

    #
    # FLIRT Signatures
    #

    def _init_flirt_signatures(self):
        if Conf.flirt_signatures_root:
            # if it's a relative path, it's relative to the angr-management package
            if os.path.isabs(Conf.flirt_signatures_root):
                flirt_signatures_root = Conf.flirt_signatures_root
            else:
                if is_pyinstaller():
                    flirt_signatures_root = os.path.join(app_root(), Conf.flirt_signatures_root)
                else:
                    # when running as a Python package, we should use the git submodule, which is on the same level
                    # with (instead of inside) the angrmanagement module directory.
                    flirt_signatures_root = os.path.join(app_root(), "..", Conf.flirt_signatures_root)
            flirt_signatures_root = os.path.normpath(flirt_signatures_root)
            _l.info("Loading FLIRT signatures from %s.", flirt_signatures_root)
            angr.flirt.load_signatures(flirt_signatures_root)

    #
    # Library docs
    #

    def _init_library_docs(self):
        GlobalInfo.library_docs = LibraryDocs()
        if Conf.library_docs_root:
            GlobalInfo.library_docs.load_func_docs(Conf.library_docs_root)

    #
    # Daemon
    #

    def _run_daemon(self, use_daemon=None):
        if use_daemon is None:
            # Load it from the configuration file
            use_daemon = Conf.use_daemon

        if not use_daemon:
            return

        # connect to daemon (if there is one)
        if not daemon_exists():
            print("[+] Starting a new daemon.")
            run_daemon_process()
            time.sleep(0.2)
        else:
            print("[+] Connecting to an existing angr management daemon.")

        while True:
            try:
                GlobalInfo.daemon_conn = daemon_conn(service=ClientService)
            except ConnectionRefusedError:
                print("[-] Connection failed... try again.")
                time.sleep(0.4)
                continue
            print("[+] Connected to daemon.")
            break

        from rpyc import BgServingThread  # pylint:disable=import-outside-toplevel

        _ = BgServingThread(GlobalInfo.daemon_conn)

    #
    # URL scheme handler setup
    #

    def _init_url_scheme_handler(self):
        if "CI" in os.environ:
            return

        # URL scheme
        from angrmanagement.logic.url_scheme import AngrUrlScheme  # pylint:disable=import-outside-toplevel

        scheme = AngrUrlScheme()
        registered, _ = scheme.is_url_scheme_registered()
        supported = scheme.is_url_scheme_supported()

        if not registered and supported and not Conf.prompted_for_url_scheme_registration:
            btn = QMessageBox.question(
                None,
                "Setting up angr URL scheme",
                'angr URL scheme allows "deep linking" from browsers and other applications '
                "by registering the angr:// protocol to the current user. Do you want to "
                "register it? You may unregister at any "
                "time in Preferences.",
                defaultButton=QMessageBox.Yes,
            )
            if btn == QMessageBox.Yes:
                try:
                    AngrUrlScheme().register_url_scheme()
                except (ValueError, FileNotFoundError) as ex:
                    QMessageBox.warning(
                        None,
                        "Error in registering angr URL scheme",
                        "Failed to register the angr URL scheme.\nThe following exception occurred:\n" + str(ex),
                    )
                    return

            Conf.prompted_for_url_scheme_registration = True
            save_config()

    #
    # Commands
    #

    def _register_commands(self):
        """
        Register basic window commands.
        """
        self.workspace.command_manager.register_commands(
            [
                BasicCommand(action.__name__, caption, action)
                for caption, action in [
                    ("Analyze: Decompile", self.decompile_current_function),
                    ("Analyze: Interact", self.interact),
                    ("Analyze: Run Analysis...", self.run_analysis),
                    ("File: Exit", self.quit),
                    ("File: Load a new binary...", self.open_file_button),
                    ("File: Load a new docker target...", self.open_docker_button),
                    ("File: Load a new trace...", self.load_trace),
                    ("File: Load angr database...", self.load_database),
                    ("File: Preferences...", self.preferences),
                    ("File: Save angr database as...", self.save_database_as),
                    ("File: Save angr database...", self.save_database),
                    ("File: Save patched binary as...", self.save_patched_binary_as),
                    ("Help: About", self.open_about_dialog),
                    ("Help: Documentation", self.open_doc_link),
                    ("View: Breakpoints", self.workspace.show_breakpoints_view),
                    ("View: Call Explorer", self.workspace.show_call_explorer_view),
                    ("View: Console", self.workspace.show_console_view),
                    ("View: Disassembly (Graph)", self.workspace.show_graph_disassembly_view),
                    ("View: Disassembly (Linear)", self.workspace.show_linear_disassembly_view),
                    ("View: Functions", self.workspace.show_functions_view),
                    ("View: Hex", self.workspace.show_hex_view),
                    ("View: Interaction", self.workspace.show_interaction_view),
                    ("View: Log", self.workspace.show_log_view),
                    ("View: New Disassembly (Graph)", self.workspace.create_and_show_graph_disassembly_view),
                    ("View: New Disassembly (Linear)", self.workspace.create_and_show_linear_disassembly_view),
                    ("View: New Hex", self.workspace.create_and_show_hex_view),
                    ("View: Patches", self.workspace.show_patches_view),
                    ("View: Proximity", self.workspace.view_proximity_for_current_function),
                    ("View: Pseudocode", self.workspace.show_pseudocode_view),
                    ("View: Registers", self.workspace.show_registers_view),
                    ("View: Stack", self.workspace.show_stack_view),
                    ("View: States", self.workspace.show_states_view),
                    ("View: Strings", self.workspace.show_strings_view),
                    ("View: Symbolic Execution", self.workspace.show_symexec_view),
                    ("View: Trace map", self.workspace.show_trace_map_view),
                    ("View: Traces", self.workspace.show_traces_view),
                    ("View: Types", self.workspace.show_types_view),
                ]
            ]
        )

    #
    # Event
    #

    def closeEvent(self, event):
        # Ask if the user wants to save things
        if (
            self.workspace.main_instance is not None
            and not self.workspace.main_instance.project.am_none
            and self.shown_at_start
        ):
            msgbox = QMessageBox()
            msgbox.setWindowTitle("Save database")
            msgbox.setText("angr management is about to exit. Do you want to save the database?")
            msgbox.setIcon(QMessageBox.Question)
            msgbox.setWindowIcon(self.windowIcon())
            msgbox.setStandardButtons(QMessageBox.Yes | QMessageBox.No | QMessageBox.Cancel)
            msgbox.setDefaultButton(QMessageBox.Yes)
            r = msgbox.exec_()

            if r == QMessageBox.Cancel:
                event.ignore()
                return
            elif r == QMessageBox.Yes:
                save_r = self.save_database()
                if not save_r:
                    # failed to save the database
                    event.ignore()
                    return

        for plugin in list(self.workspace.plugins.active_plugins.values()):
            self.workspace.plugins.deactivate_plugin(plugin)
        event.accept()

    def event(self, event):
        if event.type() == QEvent.User:
            # our event callback

            try:
                event.result = event.execute()
            except Exception as ex:  # pylint:disable=broad-except
                event.exception = ex
            event.event.set()

            return True

        return super().event(event)

    def on_screen_changed(self, screen):
        """
        When changing from one screen to another, ask disassembly views to refresh in case the DPI is changed.
        """
        self.workspace.current_screen.am_obj = screen
        self.workspace.current_screen.am_event()

    #
    # Actions
    #

    def reload(self):
        self.workspace.reload()

    def open_file_button(self):
        file_path = self.open_mainfile_dialog()
        if not file_path:
            return
        self.load_file(file_path)

    def open_trace_file_button(self):
        file_path, _ = QFileDialog.getOpenFileName(
            self,
            "Open a trace file",
            Conf.last_used_directory,
            "All files (*);;Trace files (*.trace);;",
        )
        Conf.last_used_directory = os.path.dirname(file_path)
        if not file_path:
            return
        self.load_trace_file(file_path)

    def open_docker_button(self):
        required = {
            "archr: git clone https://github.com/angr/archr && cd archr && pip install -e .": archr,
            "keystone: pip install --no-binary keystone-engine keystone-engine": keystone,
        }
        is_missing = [key for key, value in required.items() if value is None]
        if len(is_missing) > 0:
            req_msg = "You need to install the following:\n\n\t" + "\n\t".join(is_missing)
            req_msg += "\n\nInstall them to enable this functionality."
            req_msg += "\nRelaunch angr-management after install."
            QMessageBox(self).critical(None, "Dependency error", req_msg)
            return

        img_name = self._pick_image_dialog()
        if img_name is None:
            return
        target = archr.targets.DockerImageTarget(img_name, target_path=None)
        self.workspace.main_instance.add_job(LoadTargetJob(target))
        self.workspace.main_instance.img_name = img_name

    def load_trace_file(self, file_path):
        if isurl(file_path):
            QMessageBox.critical(
                self,
                "Unsupported Action",
                "Downloading trace files is not yet supported. Please specify a path to a file on disk.",
            )
        else:
            # File
            if os.path.isfile(file_path):
                try:
                    with open(file_path, "rb") as f:
                        loaded_sim_state = pickle.load(f)
                        analysis_params = {
                            "end_state": loaded_sim_state,
                            "start_addr": None,
                            "end_addr": None,
                            "block_addrs": None,
                        }
                        self.workspace.view_data_dependency_graph(analysis_params)
                    self._recent_file(file_path)
                except pickle.PickleError:
                    QMessageBox.critical(
                        self, "Unable to load trace file", "Trace file must contain a serialized SimState."
                    )
            else:
                QMessageBox.critical(
                    self,
                    "File not found",
                    f"angr management cannot open file {file_path}. Please make sure that the file exists.",
                )

    def load_file(self, file_path):
        if not isurl(file_path):
            # file
            if os.path.isfile(file_path):
                if file_path.endswith(".trace"):
                    self.workspace.load_trace_from_path(file_path)
                    return

                self.workspace.main_instance.binary_path = file_path
                self.workspace.main_instance.original_binary_path = file_path
                if file_path.endswith(".adb"):
                    self._load_database(file_path)
                else:
                    self._recent_file(file_path)
                    self.workspace.main_instance.add_job(LoadBinaryJob(file_path))
            else:
                QMessageBox.critical(
                    self,
                    "File not found",
                    f"angr management cannot open file {file_path}. Please make sure that the file exists.",
                )
        else:
            # url
            r = QMessageBox.question(
                self,
                "Downloading a file",
                f"Do you want to download a file from {file_path} and open it in angr management?",
                defaultButton=QMessageBox.Yes,
            )
            if r == QMessageBox.Yes:
                try:
                    target_path = download_url(file_path, parent=self, to_file=True, file_path=None)
                except InvalidURLError:
                    QMessageBox.critical(
                        self, "Downloading failed", "angr management failed to download the file. The URL is invalid."
                    )
                    return
                except UnexpectedStatusCodeError as ex:
                    QMessageBox.critical(
                        self,
                        "Downloading failed",
                        "angr management failed to retrieve the header of the file. "
                        f"The HTTP request returned an unexpected status code {ex.status_code}.",
                    )
                    return

                if target_path:
                    # open the file - now it's a local file
                    self.workspace.main_instance.binary_path = target_path
                    self.workspace.main_instance.original_binary_path = file_path
                    self.load_file(target_path)

    def load_database(self):
        # Open File window
        file_path, _ = QFileDialog.getOpenFileName(
            self,
            "Load angr database",
            ".",
            "angr databases (*.adb)",
        )

        if not file_path:
            return

        self._load_database(file_path)

    def save_database(self):
        if self.workspace.main_instance is None or self.workspace.main_instance.project.am_none:
            return True

        if self.workspace.main_instance.database_path is None:
            return self.save_database_as()
        else:
            return self._save_database(self.workspace.main_instance.database_path)

    def save_database_as(self):
        if self.workspace.main_instance is None or self.workspace.main_instance.project.am_none:
            return False

        default_database_path = self.workspace.main_instance.database_path
        if default_database_path is None:
            default_database_path = os.path.normpath(self.workspace.main_instance.project.filename) + ".adb"

        # Open File window
        file_path, _ = QFileDialog.getSaveFileName(
            self,
            "Save angr database",
            default_database_path,
            "angr databases (*.adb)",
        )

        if not file_path:
            return False

        if not file_path.endswith(".adb"):
            file_path = file_path + ".adb"

        return self._save_database(file_path)

    def save_patched_binary_as(self):
        if self.workspace.main_instance is None or self.workspace.main_instance.project.am_none:
            return

        file_path, _ = QFileDialog.getSaveFileName(
            self,
            "Save patched binary as...",
            self.workspace.main_instance.project.loader.main_object.binary
            + ".patched",  # FIXME: this will not work if we are loading from an angrdb
            "Any file (*)",
        )

        if file_path:
            b = self.workspace.main_instance.project.kb.patches.apply_patches_to_binary()
            with open(file_path, "wb") as f:
                f.write(b)

    def load_trace(self):
        file_path, _ = QFileDialog.getOpenFileName(self, "Load trace", ".", "bintrace (*.trace)")
        if not file_path:
            return
        self.workspace.load_trace_from_path(file_path)

    def preferences(self):
        pref = Preferences(self.workspace, parent=self)
        pref.exec_()

    def quit(self):
        self.close()

    def run_variable_recovery(self):
        self.workspace._get_or_create_view("disassembly", DisassemblyView).variable_recovery_flavor = "accurate"

    def run_induction_variable_analysis(self):
        self.workspace._get_or_create_view("disassembly", DisassemblyView).run_induction_variable_analysis()

    def run_dependency_analysis(self, func_addr: Optional[int] = None, func_arg_idx: Optional[int] = None):
        if self.workspace is None or self.workspace.main_instance is None:
            return
        dep_analysis_job = DependencyAnalysisJob(func_addr=func_addr, func_arg_idx=func_arg_idx)
        self.workspace.main_instance.add_job(dep_analysis_job)

    def run_analysis(self):
        if self.workspace:
            self.workspace.run_analysis()

    def decompile_current_function(self):
        if self.workspace is not None:
            self.workspace.decompile_current_function()

    def view_proximity_for_current_function(self):
        if self.workspace is not None:
            self.workspace.view_proximity_for_current_function()

    def interact(self):
        self.workspace.interact_program(self.workspace.main_instance.img_name)

    def show_command_palette(self, parent=None):
        dlg = CommandPaletteDialog(self.workspace, parent=(parent or self))
        dlg.setModal(True)
        dlg.exec_()
        if dlg.selected_item:
            dlg.selected_item.run()

    def show_goto_palette(self, parent=None):
        dlg = GotoPaletteDialog(self.workspace, parent=(parent or self))
        dlg.setModal(True)
        dlg.exec_()
        if dlg.selected_item:
            self.workspace.jump_to(dlg.selected_item.addr)

    #
    # Other public methods
    #

    def progress(self, status: str, progress: float, reset_stopwatch: bool = False):
        self._progress_message = status
        self._progress_percentage = progress

        if reset_stopwatch:
            self._progress_stopwatch_start_time = time.time()
        if not self._progress_update_timer.isActive():
            self._progress_update_timer.start()

        self._refresh_progress_progress_message()

    def _refresh_progress_progress_message(self):
        self._status_label.setText(self._progress_message)
        self._status_label.show()
        self._progress_label.setText(f"{self._progress_percentage:.1f}%")
        self._progress_label.show()
        self._progress_bar.setValue(round(self._progress_percentage))
        self._progress_bar.show()
        elapsed_seconds = int(time.time() - self._progress_stopwatch_start_time)
        if elapsed_seconds > 5:
            self._stopwatch_label.setText(str(datetime.timedelta(seconds=elapsed_seconds)))
            self._stopwatch_label.show()
        self._interrupt_job_button.show()
        self._progress_dialog.setLabelText(self._progress_message)
        self._progress_dialog.setValue(round(self._progress_percentage))

    def _on_progress_update_timer_timeout(self):
        self._refresh_progress_progress_message()

    def progress_done(self):
        self._progress_update_timer.stop()
        self._stopwatch_label.hide()
        self._status_label.setText("")
        self._status_label.hide()
        self._progress_label.hide()
        self._progress_bar.hide()
        self._interrupt_job_button.hide()
        self._progress_dialog.hide()

    def bring_to_front(self):
        self.setWindowState((self.windowState() & ~Qt.WindowMinimized) | Qt.WindowActive)
        self.activateWindow()
        self.raise_()

    #
    # Private methods
    #

    def _recent_file(self, file_path):
        file_path = os.path.abspath(file_path)
        self._file_menu.add_recent(file_path)
        Conf.recent_file(file_path)
        save_config()

    def _get_recent_dir(self) -> str:
        if Conf.recent_files:
            recent_dir = os.path.dirname(Conf.recent_files[-1])
            if os.path.isdir(recent_dir):
                return recent_dir
        return ""

    def _load_database(self, file_path: str):
        other_kbs = {}
        extra_info = {}

        job = LoadAngrDBJob(file_path, ["global", "pseudocode_variable_kb"], other_kbs=other_kbs, extra_info=extra_info)
        job._on_finish = partial(self._on_load_database_finished, job)
        self.workspace.main_instance.add_job(job)

    def _on_load_database_finished(self, job: LoadAngrDBJob):
        proj = job.project

        if proj is None:
            return

        self._recent_file(job.file_path)

        cfg = proj.kb.cfgs["CFGFast"]
        cfb = proj.analyses.CFB()  # it will load functions from kb

        self.workspace.main_instance.database_path = job.file_path

        self.workspace.main_instance._reset_containers()
        self.workspace.main_instance.project = proj
        self.workspace.main_instance.cfg = cfg
        self.workspace.main_instance.cfb = cfb
        if "pseudocode_variable_kb" in job.other_kbs:
            self.workspace.main_instance.pseudocode_variable_kb = job.other_kbs["pseudocode_variable_kb"]
        else:
            self.workspace.main_instance.initialize_pseudocode_variable_kb()
        self.workspace.main_instance.project.am_event(initialized=True)

        # trigger callbacks
        self.workspace.reload()
        self.workspace.main_instance.on_cfg_generated()
        self.workspace.plugins.angrdb_load_entries(job.extra_info)

    def _save_database(self, file_path):
        if self.workspace.main_instance is None or self.workspace.main_instance.project.am_none:
            return False

        self.workspace.plugins.handle_project_save(file_path)

        angrdb = AngrDB(project=self.workspace.main_instance.project)
        extra_info = self.workspace.plugins.angrdb_store_entries()
        angrdb.dump(
            file_path,
            kbs=[
                self.workspace.main_instance.kb,
                self.workspace.main_instance.pseudocode_variable_kb,
            ],
            extra_info=extra_info,
        )

        self.workspace.main_instance.database_path = file_path
        return True

    def _raise_view(self, idx: int):
        """
        Raise idx'th view in the dock manager
        """
        try:
            dock = self.dock_manager.dockWidgets()[idx]
        except IndexError:
            return
        dock.raise_()



from typing import TYPE_CHECKING, Any, Optional, Sequence, Tuple

import ailment

try:
    import pypcode
except ImportError:
    pypcode = None

import pyvex
from PySide6.QtCore import QObject, QPointF, QRectF, Qt
from PySide6.QtGui import QFont, QMouseEvent, QPainter, QTextCharFormat, QTextCursor, QTextDocument
from PySide6.QtWidgets import QGraphicsSimpleTextItem

from angrmanagement.config import Conf, ConfigurationManager
from angrmanagement.utils import string_at_addr

from .qgraph_object import QCachedGraphicsItem

if TYPE_CHECKING:
    from archinfo import RegisterOffset, TmpVar

    from angrmanagement.logic.disassembly.info_dock import InfoDock
    from angrmanagement.ui.widgets.qdisasm_base_control import QDisassemblyBaseControl


class QBlockCodeOptions:
    """
    Various options to control display of QBlockCodeObj's
    """

    show_conditional_jump_targets: bool = True
    show_variables: bool = True
    show_variable_identifiers: bool = True


class QBlockCodeObj(QObject):
    """
    Renders a generic "code" object and handles display related events.
    Instances of this class mirror an AST structure, with references in the
    `obj` property to whatever object should be displayed. Leaf nodes will add
    text to the display document during render, which is handled at the top
    level by `QBlockCode`.
    """

    obj: Any
    infodock: "InfoDock"
    parent: Any
    options: QBlockCodeOptions
    span: Optional[Tuple[int, int]]
    subobjs: Sequence["QBlockCodeObj"]
    _fmt_current: QTextCharFormat

    def __init__(self, obj: Any, infodock: "InfoDock", parent: Any, options: QBlockCodeOptions = None):
        super().__init__()
        self.obj = obj
        self.infodock = infodock
        self.parent = parent
        self.options = options or QBlockCodeOptions()
        self.span = None
        self.subobjs = []
        self._fmt_current = None
        self.update_style()
        self.create_subobjs(obj)

    @staticmethod
    def fmt() -> QTextCharFormat:
        """
        Get text char formatting for this object
        """
        fmt = QTextCharFormat()
        fmt.setForeground(Conf.disasm_view_node_mnemonic_color)
        return fmt

    def update_style(self):
        """
        Updates current rendering style before draw
        """
        self._fmt_current = self.fmt()
        if self.should_highlight():
            self._fmt_current.setBackground(Conf.disasm_view_operand_highlight_color)
            self._fmt_current.setFontWeight(QFont.Bold)

    def should_highlight(self) -> bool:
        """
        Determine whether this object should be drawn with highlight
        """
        selected = self.infodock.selected_qblock_code_obj
        return (selected is not None) and (selected is self or selected.obj is self.obj)

    def create_subobjs(self, obj):
        """
        Initialize any display subobjects for this object
        """

    def recreate_subobjs(self):
        self.subobjs.clear()
        self.create_subobjs(self.obj)

    def update(self):
        """
        Update self and parent objects
        """
        self.parent.update()

    def render_to_doc(self, cursor):
        """
        Add each subobject to the document
        """
        self.update_style()
        self.recreate_subobjs()
        span_min = cursor.position()
        for obj in self.subobjs:
            if isinstance(obj, str):
                cursor.insertText(obj, self._fmt_current)
            else:
                obj.render_to_doc(cursor)
        span_max = cursor.position()
        self.span = (span_min, span_max)

    def hit_test(self, pos: int) -> bool:
        """
        Determine whether a character offset falls within the span of this object
        """
        return self.span[0] <= pos < self.span[1]

    def get_hit_obj(self, pos: int) -> "QBlockCodeObj":
        """
        Find the leaf node for a given character offset
        """
        if not self.hit_test(pos):
            return None
        for obj in self.subobjs:
            if not isinstance(obj, str):
                hit = obj.get_hit_obj(pos)
                if hit is not None:
                    return hit
        return self

    def _add_subobj(self, obj: "QBlockCodeObj"):
        """
        Add display object `obj` to the list of subobjects
        """
        self.subobjs.append(obj)

    def add_text(self, text: str):
        """
        Add a text leaf
        """
        self._add_subobj(text)

    def add_variable(self, var):
        self._add_subobj(QVariableObj(var, self.infodock, parent=self, options=self.options))

    def mousePressEvent(self, event: QMouseEvent):  # pylint: disable=unused-argument
        self.infodock.select_qblock_code_obj(self)
        if event.button() == Qt.RightButton:
            self.infodock.disasm_view.show_context_menu_for_selected_object()

    def mouseDoubleClickEvent(self, event: QMouseEvent):
        pass

    @property
    def should_highlight_line(self):
        return any(obj.should_highlight_line for obj in self.subobjs if isinstance(obj, QBlockCodeObj))


class QVariableObj(QBlockCodeObj):
    """
    Renders a variable
    """

    @staticmethod
    def fmt() -> QTextCharFormat:
        fmt = QTextCharFormat()
        fmt.setForeground(Conf.disasm_view_variable_label_color)
        return fmt

    def create_subobjs(self, obj):
        ident = "<%s>" % (obj.ident if obj.ident else "") if self.options.show_variable_identifiers else ""
        self.add_text(obj.name + ident)


class QAilObj(QBlockCodeObj):
    """
    Renders an AIL object
    """

    def __init__(self, obj: Any, instance, *args, stmt=None, **kwargs):
        self.stmt = stmt or obj
        self.instance = instance
        super().__init__(obj, *args, **kwargs)

    def create_subobjs(self, obj: Any):
        self.add_ailobj(obj)

    def add_ailobj(self, obj: Any):
        """
        Map appropriate AIL type to the display type
        """
        subobjcls = {
            ailment.statement.Assignment: QAilAssignmentObj,
            ailment.statement.Store: QAilStoreObj,
            ailment.statement.Jump: QAilJumpObj,
            ailment.statement.ConditionalJump: QAilConditionalJumpObj,
            ailment.statement.Return: QAilReturnObj,
            ailment.statement.Call: QAilCallObj,
            ailment.expression.Const: QAilConstObj,
            ailment.expression.Tmp: QAilTmpObj,
            ailment.expression.Register: QAilRegisterObj,
            ailment.expression.UnaryOp: QAilUnaryOpObj,
            ailment.expression.BinaryOp: QAilBinaryOpObj,
            ailment.expression.Convert: QAilConvertObj,
            ailment.expression.Load: QAilLoadObj,
        }.get(type(obj), QAilTextObj)
        subobj = subobjcls(obj, self.instance, self.infodock, parent=self, options=self.options, stmt=self.stmt)
        self._add_subobj(subobj)

    @property
    def should_highlight_line(self):
        ail_obj_ins_addr = getattr(self.obj, "ins_addr", None)
        if ail_obj_ins_addr is not None and self.infodock.is_instruction_selected(ail_obj_ins_addr):
            return True
        return super().should_highlight_line

    def mousePressEvent(self, event: QMouseEvent):  # pylint: disable=unused-argument
        super().mousePressEvent(event)
        button = event.button()
        if button == Qt.LeftButton:
            ail_obj_ins_addr = getattr(self.obj, "ins_addr", None)
            if ail_obj_ins_addr is not None:
                self.infodock.select_instruction(ail_obj_ins_addr)


class QAilTextObj(QAilObj):
    """
    Renders an AIL object via __str__
    """

    def create_subobjs(self, obj: Any):
        self.add_text(str(obj))


class QAilAssignmentObj(QAilTextObj):
    """
    Renders an ailment.statement.Assignment
    """

    def create_subobjs(self, obj: ailment.statement.Assignment):
        self.add_ailobj(obj.dst)
        self.add_text(" = ")
        self.add_ailobj(obj.src)


class QAilStoreObj(QAilTextObj):
    """
    Renders an ailment.statement.Store
    """

    def create_subobjs(self, obj: ailment.statement.Store):
        if obj.variable is None or not self.options.show_variables:
            self.add_text("*(")
            self.add_ailobj(obj.addr)
            self.add_text(") = ")
            self.add_ailobj(obj.data)
        else:
            self.add_variable(obj.variable)
            self.add_text(" = ")
            self.add_ailobj(obj.data)


class QAilJumpObj(QAilTextObj):
    """
    Renders an ailment.statement.Jump
    """

    def create_subobjs(self, obj: ailment.statement.Jump):
        self.add_text("goto ")
        self.add_ailobj(obj.target)


class QAilConditionalJumpObj(QAilTextObj):
    """
    Renders an ailment.statement.ConditionalJump
    """

    def create_subobjs(self, obj: ailment.statement.ConditionalJump):
        self.add_text("if ")
        self.add_ailobj(obj.condition)

        if self.options.show_conditional_jump_targets:
            self.add_text(" goto ")
            self.add_ailobj(obj.true_target)
            self.add_text(" else goto ")
            self.add_ailobj(obj.false_target)


class QAilReturnObj(QAilTextObj):
    """
    Renders an ailment.statement.Return
    """

    def create_subobjs(self, obj: ailment.statement.Return):
        self.add_text("return ")
        for expr in obj.ret_exprs:
            self.add_ailobj(expr)


class QAilCallObj(QAilTextObj):
    """
    Renders an ailment.statement.Call
    """

    def create_subobjs(self, obj: ailment.statement.Call):
        if obj.ret_expr is not None and self.stmt is self.obj:
            self.add_ailobj(obj.ret_expr)
            self.add_text(" = ")
        self.add_ailobj(obj.target)
        self.add_text("(")
        if obj.args:
            for i, arg in enumerate(obj.args):
                if i > 0:
                    self.add_text(", ")
                self.add_ailobj(arg)
        self.add_text(")")


class QAilConstObj(QAilTextObj):
    """
    Renders an ailment.expression.Const
    """

    @staticmethod
    def fmt() -> QTextCharFormat:
        fmt = QTextCharFormat()
        fmt.setForeground(Conf.disasm_view_operand_constant_color)
        return fmt

    def create_subobjs(self, obj: ailment.expression.Const):
        # take care of labels first
        kb = self.infodock.disasm_view.disasm.kb
        if obj.value in kb.labels:
            self.add_text(kb.labels[obj.value])
            return

        data_str = string_at_addr(
            self.instance.cfg,
            obj.value,
            self.instance.project,
        )
        if data_str:
            self.add_text(data_str)
        else:
            self.add_text(f"{obj.value:#x}")

    def should_highlight(self) -> bool:
        return (
            isinstance(self.infodock.selected_qblock_code_obj, QAilConstObj)
            and self.infodock.selected_qblock_code_obj.obj.value == self.obj.value
        )

    def mouseDoubleClickEvent(self, event: QMouseEvent):
        super().mouseDoubleClickEvent(event)
        button = event.button()
        if button == Qt.LeftButton:
            src_ins_addr = getattr(self.stmt, "ins_addr", None)
            self.infodock.disasm_view.jump_to(self.obj.value, src_ins_addr=src_ins_addr, use_animation=True)


class QAilTmpObj(QAilTextObj):
    """
    Renders an ailment.expression.Tmp
    """

    @staticmethod
    def fmt() -> QTextCharFormat:
        fmt = QTextCharFormat()
        fmt.setForeground(Conf.disasm_view_operand_color)
        return fmt


class QAilRegisterObj(QAilTextObj):
    """
    Renders an ailment.expression.Register
    """

    @staticmethod
    def fmt() -> QTextCharFormat:
        fmt = QTextCharFormat()
        fmt.setForeground(Conf.disasm_view_operand_color)
        return fmt

    def create_subobjs(self, obj: ailment.expression.Register):
        if obj.variable is not None and self.options.show_variables:
            self.add_variable(obj.variable)
        else:
            s = f"{obj.reg_name}" if hasattr(obj, "reg_name") else "reg_%d<%d>" % (obj.reg_offset, obj.bits // 8)
            self.add_text(s)

    def should_highlight(self) -> bool:
        sel = self.infodock.selected_qblock_code_obj
        return isinstance(sel, QAilRegisterObj) and sel.obj == self.obj


class QAilUnaryOpObj(QAilTextObj):
    """
    Renders an ailment.expression.UnaryOp
    """

    def create_subobjs(self, obj: ailment.expression.UnaryOp):
        self.add_text("(")
        self.add_text(obj.op + " ")
        self.add_ailobj(obj.operand)
        self.add_text(")")


class QAilBinaryOpObj(QAilTextObj):
    """
    Renders an ailment.expression.BinaryOp
    """

    def create_subobjs(self, obj: ailment.expression.BinaryOp):
        self.add_text("(")
        self.add_ailobj(obj.operands[0])
        verbose_op = obj.OPSTR_MAP.get(obj.verbose_op, obj.verbose_op)
        if verbose_op is None:
            verbose_op = "unknown_op"
        self.add_text(" " + verbose_op + " ")
        self.add_ailobj(obj.operands[1])
        self.add_text(")")


class QAilConvertObj(QAilTextObj):
    """
    Renders an ailment.expression.Convert
    """

    def create_subobjs(self, obj: ailment.expression.Convert):
        self.add_text("Conv(%d->%d, " % (obj.from_bits, obj.to_bits))
        self.add_ailobj(obj.operand)
        self.add_text(")")


class QAilLoadObj(QAilTextObj):
    """
    Renders an ailment.expression.Load
    """

    def create_subobjs(self, obj: ailment.expression.Load):
        if obj.variable is not None and self.options.show_variables:
            self.add_variable(obj.variable)
        else:
            self.add_text("*(")
            self.add_ailobj(obj.addr)
            self.add_text(")")


OBJ_CLASS_TO_QBLOCKCODE_CLASS = {}


class QIROpObj(QBlockCodeObj):
    """
    Renders a Lifter IR object
    """

    @staticmethod
    def fmt() -> QTextCharFormat:
        fmt = QTextCharFormat()
        fmt.setForeground(Conf.disasm_view_ir_default_color)
        return fmt

    def __init__(self, obj: Any, *args, irobj=None, **kwargs):
        self.irobj = irobj or obj
        super().__init__(obj, *args, **kwargs)

    def create_subobjs(self, obj):
        self.add_irobj(obj.obj)

    def add_irobj(self, obj):
        subobjcls = OBJ_CLASS_TO_QBLOCKCODE_CLASS.get(type(obj), QIROpTextObj)
        subobj = subobjcls(obj, self.infodock, parent=self, options=self.options, irobj=self.irobj)
        self._add_subobj(subobj)


class QIROpTextObj(QIROpObj):
    """
    Renders a Lifter IR object using the object's __str__, or as hexadecimal
    if an integer type.
    """

    def create_subobjs(self, obj: Any):
        if isinstance(obj, int):
            self.add_text("%#x" % obj)
        else:
            self.add_text(str(obj))


class QIrOpPcodeOp(QIROpTextObj):
    """
    Renders a P-code op.
    """

    def create_subobjs(self, obj: "pypcode.PcodeOp"):
        self.add_text(pypcode.PcodePrettyPrinter.fmt_op(obj))


if pypcode:
    OBJ_CLASS_TO_QBLOCKCODE_CLASS.update({pypcode.PcodeOp: QIrOpPcodeOp})


class QIROpVexConstObj(QIROpTextObj):
    """
    Renders a pyvex.expr.Const
    """

    @staticmethod
    def fmt() -> QTextCharFormat:
        fmt = QTextCharFormat()
        fmt.setForeground(Conf.disasm_view_operand_constant_color)
        return fmt

    def should_highlight(self) -> bool:
        return (
            isinstance(self.infodock.selected_qblock_code_obj, QIROpVexConstObj)
            and self.infodock.selected_qblock_code_obj.obj._con == self.obj._con
        )


class VexIRTmpWrapper:
    """
    A wrapper class for VEX temps
    """

    __slots__ = (
        "tid",
        "reg_name",
    )

    tid: "TmpVar"
    reg_name: Optional[str]

    def __init__(self, tid: "TmpVar", reg_name: Optional[str] = None):
        self.tid = tid
        self.reg_name = reg_name or ("t%d" % self.tid)

    def __str__(self):
        return self.reg_name


class VexIRRegWrapper:
    """
    A wrapper class for VEX registers
    """

    __slots__ = (
        "offset",
        "reg_name",
    )

    offset: "RegisterOffset"
    reg_name: Optional[str]

    def __init__(self, offset: "RegisterOffset", reg_name: Optional[str] = None):
        self.offset = offset
        self.reg_name = reg_name or ("offset=%s" % self.offset)

    def __str__(self):
        return self.reg_name


class QIROpVexWrTmpObj(QIROpTextObj):
    """
    Renders a pyvex.stmt.WrTmp
    """

    def create_subobjs(self, obj: pyvex.stmt.WrTmp):
        irsb = self.irobj.irsb
        self.add_irobj(VexIRTmpWrapper(obj.tmp))
        self.add_text(" = ")
        if isinstance(obj.data, pyvex.expr.Get):
            reg_name = irsb.arch.translate_register_name(obj.data.offset, obj.data.result_size(irsb.tyenv) // 8)
            self.add_irobj(VexIRRegWrapper(obj.data, reg_name))
        else:
            self.add_irobj(obj.data)


class QIROpVexRdTmpObj(QIROpTextObj):
    """
    Renders a pyvex.expr.RdTmp
    """

    def create_subobjs(self, obj: pyvex.expr.RdTmp):
        self.add_irobj(VexIRTmpWrapper(obj.tmp))


class QIROpVexTmpObj(QIROpTextObj):
    """
    Renders a VEX temporary
    """

    @staticmethod
    def fmt() -> QTextCharFormat:
        fmt = QTextCharFormat()
        fmt.setForeground(Conf.disasm_view_operand_color)
        return fmt

    def should_highlight(self) -> bool:
        return (
            isinstance(self.infodock.selected_qblock_code_obj, QIROpVexTmpObj)
            and self.infodock.selected_qblock_code_obj.obj.tid == self.obj.tid
        )


class QIROpVexRegObj(QIROpTextObj):
    """
    Renders a VEX register
    """

    @staticmethod
    def fmt() -> QTextCharFormat:
        fmt = QTextCharFormat()
        fmt.setForeground(Conf.disasm_view_operand_color)
        return fmt

    def should_highlight(self) -> bool:
        return (
            isinstance(self.infodock.selected_qblock_code_obj, QIROpVexRegObj)
            and self.infodock.selected_qblock_code_obj.obj.offset == self.obj.offset
        )


class QIROpVexStoreObj(QIROpTextObj):
    """
    Renders a pyvex.stmt.Store
    """

    def create_subobjs(self, obj: pyvex.stmt.Store):
        # "ST%s(%s) = %s" % (self.endness[-2:].lower(), self.addr, self.data)
        self.add_text(f"ST{obj.endness[-2:].lower()}(")
        self.add_irobj(obj.addr)
        self.add_text(") = ")
        self.add_irobj(obj.data)


class QIROpVexLoadObj(QIROpTextObj):
    """
    Renders a pyvex.expr.Load
    """

    def create_subobjs(self, obj: pyvex.expr.Load):
        self.add_text(f"LD{obj.end[-2:].lower()}:{obj.ty[4:]}(")
        self.add_irobj(obj.addr)
        self.add_text(")")


class QIROpVexPutObj(QIROpTextObj):
    """
    Renders a pyvex.stmt.Put
    """

    def create_subobjs(self, obj: pyvex.stmt.Put):
        irsb = self.irobj.irsb
        reg_name = irsb.arch.translate_register_name(obj.offset, obj.data.result_size(irsb.tyenv) // 8)
        self.add_text("PUT(")
        self.add_irobj(VexIRRegWrapper(obj.offset, reg_name))
        self.add_text(") = ")
        self.add_irobj(obj.data)


class QIROpVexExitObj(QIROpTextObj):
    """
    Renders a pyvex.stmt.Exit
    """

    def create_subobjs(self, obj: pyvex.stmt.Exit):
        irsb = self.irobj.irsb
        arch = irsb.arch
        reg_name = arch.translate_register_name(irsb.offsIP, arch.bits // 8)
        self.add_text("if (")
        self.add_irobj(obj.guard)
        self.add_text(") { PUT(")
        self.add_irobj(VexIRRegWrapper(obj.offsIP, reg_name))
        self.add_text(") = ")
        self.add_irobj(obj.dst.value)
        self.add_text("; ")
        self.add_irobj(obj.jumpkind)
        self.add_text(" }")


class QIROpVexBinopObj(QIROpTextObj):
    """
    Renders a pyvex.expr.Binop
    """

    def create_subobjs(self, obj: pyvex.expr.Binop):
        self.add_text(obj.op[4:])
        self.add_text("(")
        self.add_irobj(obj.args[0])
        self.add_text(",")
        self.add_irobj(obj.args[1])
        self.add_text(")")


class QIROpVexUnopObj(QIROpTextObj):
    """
    Renders a pyvex.expr.Unop
    """

    def create_subobjs(self, obj: pyvex.expr.Unop):
        self.add_text(obj.op[4:])
        self.add_text("(")
        self.add_irobj(obj.args[0])
        self.add_text(")")


OBJ_CLASS_TO_QBLOCKCODE_CLASS.update(
    {
        pyvex.stmt.WrTmp: QIROpVexWrTmpObj,
        pyvex.expr.RdTmp: QIROpVexRdTmpObj,
        pyvex.stmt.Store: QIROpVexStoreObj,
        pyvex.expr.Load: QIROpVexLoadObj,
        pyvex.stmt.Put: QIROpVexPutObj,
        pyvex.stmt.Exit: QIROpVexExitObj,
        pyvex.expr.Const: QIROpVexConstObj,
        pyvex.expr.Binop: QIROpVexBinopObj,
        pyvex.expr.Unop: QIROpVexUnopObj,
        VexIRTmpWrapper: QIROpVexTmpObj,
        VexIRRegWrapper: QIROpVexRegObj,
    }
)


class QBlockCode(QCachedGraphicsItem):
    """
    Top-level code widget for a selection of text. Will construct an AST using
    QBlockCodeObj, mirroring the structure associated with the target object.
    This text is then rendered using a QTextDocument, with appropriate styles
    applied to it. Interaction events will be propagated to corresponding
    objects.
    """

    GRAPH_ADDR_SPACING = 20

    addr: int
    _addr_str: str
    obj: QBlockCodeObj
    _config: ConfigurationManager
    disasm_view: "QDisassemblyBaseControl"
    infodock: "InfoDock"
    parent: Any

    def __init__(
        self,
        addr: int,
        obj: QBlockCodeObj,
        config: ConfigurationManager,
        disasm_view: "QDisassemblyBaseControl",
        instance,
        infodock: "InfoDock",
        parent: Any = None,
    ):
        super().__init__(parent=parent)
        self.addr = addr
        self._addr_str = "%08x" % self.addr
        self._addr_item: QGraphicsSimpleTextItem = None
        self.obj = obj
        self._width = 0
        self._height = 0
        self._config = config
        self.parent = parent
        self.instance = instance
        self.infodock = infodock
        self._disasm_view = disasm_view
        self._qtextdoc = QTextDocument()
        self._qtextdoc.setDefaultFont(self._config.disasm_font)
        self._qtextdoc.setDocumentMargin(0)

        self._addr_item = QGraphicsSimpleTextItem(self._addr_str, self)
        self._addr_item.setBrush(Conf.disasm_view_node_address_color)
        self._addr_item.setFont(Conf.disasm_font)

        self.update_document()
        self.setToolTip("Address: " + self._addr_str)

        self.refresh()

    def refresh(self):
        self._addr_item.setVisible(self._disasm_view.show_address)
        self._layout_items_and_update_size()

    def update_document(self):
        self._qtextdoc.clear()
        cur = QTextCursor(self._qtextdoc)
        self.obj.render_to_doc(cur)

    def paint(self, painter, option, widget):  # pylint: disable=unused-argument
        self.update_document()
        painter.setRenderHints(QPainter.Antialiasing | QPainter.SmoothPixmapTransform)
        painter.setFont(self._config.disasm_font)

        if self.infodock.is_instruction_selected(self.addr) or self.obj.should_highlight_line:
            highlight_color = Conf.disasm_view_node_instruction_selected_background_color
            painter.setBrush(highlight_color)
            painter.setPen(highlight_color)
            painter.drawRect(0, 0, self.width, self.height)

        x = 0

        if self._disasm_view.show_address:
            x += self._addr_item.boundingRect().width() + self.GRAPH_ADDR_SPACING

        painter.translate(QPointF(x, 0))
        self._qtextdoc.drawContents(painter)

    #
    # Event handlers
    #

    def get_obj_for_mouse_event(self, event: QMouseEvent) -> QBlockCodeObj:
        p = event.pos()

        if self._disasm_view.show_address:
            offset = self._addr_item.boundingRect().width() + self.GRAPH_ADDR_SPACING
            p.setX(p.x() - offset)

        if p.x() >= 0:
            hitpos = self._qtextdoc.documentLayout().hitTest(p, Qt.HitTestAccuracy.ExactHit)
            if hitpos >= 0:
                return self.obj.get_hit_obj(hitpos)

        return None

    def mousePressEvent(self, event):
        if event.button() == Qt.LeftButton:
            self.infodock.select_instruction(self.addr)

        obj = self.get_obj_for_mouse_event(event)
        if obj is not None:
            obj.mousePressEvent(event)

    def mouseDoubleClickEvent(self, event):
        obj = self.get_obj_for_mouse_event(event)
        if obj is not None:
            obj.mouseDoubleClickEvent(event)

    #
    # Private methods
    #

    def _layout_items_and_update_size(self):
        self.update_document()

        x, y = 0, 0
        if self._disasm_view.show_address:
            self._addr_item.setPos(x, y)
            x += self._addr_item.boundingRect().width() + self.GRAPH_ADDR_SPACING

        x += self._qtextdoc.size().width()
        y += self._qtextdoc.size().height()
        self._width = x
        self._height = y
        self.recalculate_size()

    def _boundingRect(self):
        return QRectF(0, 0, self._width, self._height)
